import json

from .enum import StorageClassType
from .models import (BucketInfo, CommonPrefixInfo,
                     CompleteMultipartUploadResult, CopyObjectResult,
                     CreateMultipartUploadResult, DeletedErrInfo,
                     DeletedObjectInfo, DeleteMarkerInfo, DeleteObjectsResult,
                     GetObjectAclResult, GrantInfo, ListBucketResult,
                     ListMultipartUploadsResult, ListObjectsResult,
                     ListObjectVersionsResult, ListPartsResult,
                     MultipartUploadInfo, ObjectInfo, ObjectVersionInfo,
                     UploadPartCopyResult, UploadPartInfo, UserInfo)
from .models2 import (ListedObjectVersion,
                      ListObjectVersionsOutput, Owner)
from .utils import (get_etag, get_value, parse_modify_time_to_utc_datetime)


def convert_list_buckets_result(resp):
    result = ListBucketResult(resp)
    data = json.loads(resp.read().decode('utf-8'))
    result.owner = UserInfo(
        get_value(data['Owner'], 'ID'),
        get_value(data['Owner'], 'Name'),
    )

    bkt_list = get_value(data, 'Buckets') or []
    for bkt in bkt_list:
        result.bucket_list.append(BucketInfo(
            get_value(bkt, 'Name'),
            get_value(bkt, 'Location'),
            get_value(bkt, 'CreationDate'),
            get_value(bkt, 'ExtranetEndpoint'),
            get_value(bkt, 'IntranetEndpoint')))
    return result


def convert_create_multipart_upload_result(resp):
    result = CreateMultipartUploadResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.upload_id = get_value(data, 'UploadId')
    result.bucket = get_value(data, 'Bucket')
    result.key = get_value(data, 'Key')
    return result


def convert_list_multipart_uploads_result(resp):
    result = ListMultipartUploadsResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.bucket = get_value(data, 'Bucket')
    result.upload_id_marker = get_value(data, 'UploadIdMarker')
    result.next_key_marker = get_value(data, 'NextKeyMarker')
    result.next_upload_id_marker = get_value(data, 'NextUploadIdMarker')
    result.delimiter = get_value(data, 'Delimiter')
    result.prefix = get_value(data, 'Prefix')
    result.max_uploads = get_value(data, 'MaxUploads')
    if get_value(data, 'IsTruncated'):
        result.is_truncated = get_value(data, 'IsTruncated')
    else:
        result.is_truncated = False

    upload_list = get_value(data, 'Uploads') or []
    for upload in upload_list:
        initiated = get_value(upload, 'Initiated')
        if initiated:
            initiated = parse_modify_time_to_utc_datetime(initiated)
        multipart_upload_info = MultipartUploadInfo(
            get_value(upload, 'Key'),
            get_value(upload, 'UploadId'),
            get_value(upload, 'StorageClass'),
            initiated,
        )

        owner = get_value(upload, 'Owner')
        if owner:
            id = get_value(owner, 'ID')
            name = get_value(owner, 'DisplayName')
            multipart_upload_info.owner = UserInfo(id, name)

        result.upload_list.append(multipart_upload_info)

    common_prefix_list = get_value(data, 'CommonPrefixes') or []
    for common_prefix in common_prefix_list:
        result.common_prefix_list.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))
    return result


def convert_copy_object_result(resp):
    result = CopyObjectResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.etag = get_etag(data)
    result.last_modified = get_value(data, 'LastModified')
    if result.last_modified:
        result.last_modified = parse_modify_time_to_utc_datetime(result.last_modified)

    return result


def convert_upload_part_copy_result(resp):
    result = UploadPartCopyResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.etag = get_etag(data)
    result.last_modified = get_value(data, 'LastModified')
    if result.last_modified:
        result.last_modified = parse_modify_time_to_utc_datetime(result.last_modified)
    return result


def convert_list_parts_result(resp):
    result = ListPartsResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.bucket = get_value(data, 'Bucket')
    result.key = get_value(data, 'Key')
    result.upload_id = get_value(data, 'UploadId')
    result.part_number_marker = get_value(data, 'PartNumberMarker')
    result.next_part_number_marker = get_value(data, 'NextPartNumberMarker')
    result.max_parts = get_value(data, 'MaxParts')
    result.storage_class = get_value(data, 'StorageClass')

    if get_value(data, 'IsTruncated'):
        result.is_truncated = get_value(data, 'IsTruncated')
    else:
        result.is_truncated = False

    owner = get_value(data, 'Owner')
    if owner:
        id = get_value(owner, 'ID')
        name = get_value(owner, 'DisplayName')
        result.owner = UserInfo(id, name)

    part_list = get_value(data, 'Parts') or []
    for part in part_list:
        last_modified = get_value(part, 'LastModified')
        if last_modified:
            last_modified = parse_modify_time_to_utc_datetime(last_modified)
        result.part_list.append(UploadPartInfo(
            get_value(part, 'PartNumber'),
            last_modified,
            get_etag(part),
            get_value(part, 'Size')
        ))
    return result


def convert_complete_multipart_upload_result(resp):
    result = CompleteMultipartUploadResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.location = get_value(data, 'Location')
    result.bucket = get_value(data, 'Bucket')
    result.key = get_value(data, 'Key')
    result.etag = get_etag(data)
    return result


def convert_list_objects_result(resp):
    result = ListObjectsResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.name = get_value(data, 'Name')
    result.prefix = get_value(data, 'Prefix')
    result.marker = get_value(data, 'Marker')
    result.max_keys = get_value(data, 'MaxKeys', int)
    result.next_marker = get_value(data, 'NextMarker')
    result.delimiter = get_value(data, 'Delimiter')
    if get_value(data, 'EncodingType'):
        result.encoding_type = get_value(data, 'EncodingType')
    else:
        result.encoding_type = 'url'

    if get_value(data, 'IsTruncated'):
        result.is_truncated = get_value(data, 'IsTruncated')
    else:
        result.is_truncated = False

    common_prefix_list = get_value(data, 'CommonPrefixes') or []
    for common_prefix in common_prefix_list:
        result.common_prefix_list.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))

    object_list = get_value(data, 'Contents') or []
    for object in object_list:
        last_modified = get_value(object, 'LastModified')
        if last_modified:
            last_modified = parse_modify_time_to_utc_datetime(last_modified)
        object_info = ObjectInfo(
            get_value(object, 'Key'),
            last_modified,
            get_etag(object),
            get_value(object, 'Size'),
            get_value(object, 'StorageClass')
        )
        owner_info = get_value(object, 'Owner')
        if owner_info:
            object_info.owner = UserInfo(
                get_value(owner_info, "ID"),
                get_value(owner_info, 'DisplayName')
            )
        result.object_list.append(object_info)
    return result


def convert_list_object_versions_result(resp):
    result = ListObjectVersionsResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.name = get_value(data, 'Name')
    result.prefix = get_value(data, 'Prefix')
    result.key_marker = get_value(data, 'KeyMarker')
    result.version_id_marker = get_value(data, 'VersionIdMarker')
    result.next_key_marker = get_value(data, 'NextKeyMarker')
    result.next_version_id_marker = get_value(data, 'NextVersionIdMarker')
    result.delimiter = get_value(data, 'Delimiter')
    result.max_keys = get_value(data, 'MaxKeys', lambda x: int(x))

    if get_value(data, 'EncodingType'):
        result.encoding_type = get_value(data, 'EncodingType')
    else:
        result.encoding_type = 'url'

    if get_value(data, 'IsTruncated'):
        result.is_truncated = get_value(data, 'IsTruncated')
    else:
        result.is_truncated = False

    common_prefix_list = get_value(data, 'CommonPrefixes') or []
    for common_prefix in common_prefix_list:
        result.common_prefix_list.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))

    object_list = get_value(data, 'Versions') or []
    for object in object_list:
        last_modified = get_value(object, 'LastModified')
        if last_modified:
            last_modified = parse_modify_time_to_utc_datetime(last_modified)
        object_info = ObjectVersionInfo(
            get_value(object, 'Key'),
            get_value(object, 'IsLatest'),
            last_modified,
            get_etag(object),
            get_value(object, 'Size', lambda x: int(x)),
            get_value(object, 'StorageClass'),
            get_value(object, 'VersionId')
        )
        owner_info = get_value(object, 'Owner')
        if owner_info:
            object_info.owner = UserInfo(
                get_value(owner_info, "ID"),
                get_value(owner_info, 'DisplayName')
            )
        result.version_list.append(object_info)

    delete_marker_list = get_value(data, 'DeleteMarkers') or []
    for delete_marker in delete_marker_list:
        delete_marker_info = DeleteMarkerInfo(
            get_value(delete_marker, 'Key'),
            get_value(delete_marker, 'IsLatest'),
            get_value(delete_marker, 'LastModified'),
            get_value(delete_marker, 'VersionId')
        )
        owner_info = get_value(delete_marker, 'Owner')
        if owner_info:
            delete_marker_info.owner = UserInfo(
                get_value(owner_info, "ID"),
                get_value(owner_info, 'DisplayName')
            )
        result.delete_marker_list.append(delete_marker_info)
    return result


def convert_delete_objects_result(resp):
    result = DeleteObjectsResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    delete_list = get_value(data, 'Deleted') or []
    for delete in delete_list:
        result.deleted_list.append(DeletedObjectInfo(
            get_value(delete, 'Key'),
            get_value(delete, 'VersionId'),
            get_value(delete, 'DeleteMarker'),
            get_value(delete, 'DeleteMarkerVersionId')
        ))
    err_list = get_value(data, 'Error') or []
    for err in err_list:
        result.error_list.append(DeletedErrInfo(
            get_value(err, 'Code'),
            get_value(err, 'Message'),
            get_value(err, 'Key'),
            get_value(err, 'VersionId')
        ))
    return result


def convert_get_object_acl_result(resp):
    result = GetObjectAclResult(resp)
    data = json.loads(resp.read().decode('utf-8'))

    result.owner = UserInfo(
        get_value(data['Owner'], 'ID'),
        get_value(data['Owner'], 'DisplayName'),
    )

    grant_list = data['Grants'] or []
    for grant in grant_list:
        result.grant_list.append(GrantInfo(
            get_value(grant['Grantee'], 'ID'),
            get_value(grant['Grantee'], 'DisplayName'),
            get_value(grant['Grantee'], 'Type'),
            get_value(grant['Grantee'], 'Canned'),
            get_value(grant, 'Permission'),
        ))
    return result



# -*- coding: utf-8 -*-
import base64
import datetime
import hmac
import json
import logging
from hashlib import sha256
from urllib.parse import quote

import pytz

from .consts import DATE_FORMAT, UNSIGNED_PAYLOAD, LAST_MODIFY_TIME_DATE_FORMAT
from .credential import FederationCredentials, StaticCredentials
from .exceptions import TosClientError
from .models2 import PreSignedPostSignatureOutPut, ContentLengthRange
from .utils import to_bytes, _param_to_quoted_query

logger = logging.getLogger(__name__)


def _canonical_query_string_params(params):
    results = []
    for param in sorted(params):
        value = str(params[param])
        results.append('%s=%s' % (quote(param, safe='-_.~'),
                                  quote(value, safe='-_.~')))
    cqs = '&'.join(results)
    return cqs


def _signed_headers(headers):
    hl = sorted(headers.items(), key=lambda d: d[0].lower())
    vl = []
    for v in hl:
        vl.append(v[0].lower())
    return ';'.join(vl)


def _canonical_headers(headers):
    hl = sorted(headers.items(), key=lambda d: d[0].lower())
    s = ''
    for val in hl:
        if isinstance(val[1], list):
            tlist = sorted(val[1])
            for v in tlist:
                s += val[0] + ':' + v + '\n'
        else:
            s += val[0].lower() + ':' + str(val[1]) + '\n'
    return s


def _canonical_request(req):
    cr = [req.method.upper(), quote(req.path, safe='/~'), _canonical_query_string_params(req.params),
          _canonical_headers(req.headers), _signed_headers(req.headers)]
    if req.headers.get('x-tos-content-sha256'):
        cr.append(req.headers['x-tos-content-sha256'])
    else:
        cr.append(UNSIGNED_PAYLOAD)
    return '\n'.join(cr)


def _x_tos_policy_canonical_request(params):
    cr = [_canonical_query_string_params(params), UNSIGNED_PAYLOAD]
    return '\n'.join(cr)


def _sign(key, msg, hex=False):
    if hex:
        sig = hmac.new(key, msg.encode('utf-8'), sha256).hexdigest()
    else:
        sig = hmac.new(key, msg.encode('utf-8'), sha256).digest()
    return sig


def _check_policy_key(key):
    if not key:
        raise TosClientError('invalid preSingedCondition key')


def _get_post_policy(date: str, expire: int, algorithm, credential, bucket, object_key,
                     conditions: [], content_length_range: ContentLengthRange = None, sts: str = None) -> dict:
    time = datetime.datetime.strptime(date, DATE_FORMAT).replace(tzinfo=pytz.utc) + datetime.timedelta(seconds=expire)
    post_policy = {
        "expiration": time.strftime(LAST_MODIFY_TIME_DATE_FORMAT)
    }

    cond = [{"x-tos-algorithm": algorithm}, {"x-tos-credential": credential}, {"x-tos-date": date}, {"key": object_key},
            {"bucket": bucket}]
    if sts:
        cond.append({'x-tos-security-token': sts})

    for c in conditions:
        _check_policy_key(c.key)
        if c.operator:
            cond.append([c.operator, "${}".format(c.key), c.value])
            continue
        cond.append({c.key: c.value})

    if content_length_range:
        cond.append(["content-length-range", content_length_range.start, content_length_range.end])

    post_policy["conditions"] = cond
    return post_policy


def _get_policy(conditions: []):
    policy = {}
    cond = []
    for c in conditions:
        _check_policy_key(c.key)
        if c.operator:
            cond.append([c.operator, "${}".format(c.key), c.value])
            continue
        cond.append({c.key: c.value})
    policy['conditions'] = cond
    return policy


class AuthBase():
    def __init__(self, credentials_provider, region):
        self.credentials_provider = credentials_provider
        self.region = region.strip()
        self.credential = None

    def sign_request(self, req):
        self.credential = self.credentials_provider.get_credentials()
        if self.credential.get_security_token():
            req.headers["x-tos-security-token"] = self.credential.get_security_token()

        date = datetime.datetime.utcnow().strftime(DATE_FORMAT)
        req.headers['Date'] = date
        req.headers['x-tos-date'] = date

        signature = self._make_signature(req=req, date=date)
        req.headers['Authorization'] = self._inject_signature_to_request(req, signature, date)

    def sign_url(self, req, expires):
        if expires is None:
            expires = 60 * 60
        date = datetime.datetime.utcnow().strftime(DATE_FORMAT)
        self.credential = self.credentials_provider.get_credentials()

        req.params['X-Tos-Algorithm'] = 'TOS4-HMAC-SHA256'
        req.params['X-Tos-Credential'] = self._credential(date)
        req.params['X-Tos-Date'] = date
        req.params['X-Tos-Expires'] = expires
        req.params['X-Tos-SignedHeaders'] = _signed_headers(req.headers)

        if self.credential.get_security_token():
            req.params["X-Tos-Security-Token"] = self.credential.get_security_token()
        req.params['X-Tos-Signature'] = self._make_signature(req=req, date=date)

        return req.url + '?' + '&'.join(_param_to_quoted_query(k, v) for k, v in req.params.items())

    def post_sign(self, bucket: str, key: str, expires: int, conditions: [],
                  content_length_range: ContentLengthRange) -> PreSignedPostSignatureOutPut:
        date = datetime.datetime.utcnow().strftime(DATE_FORMAT)
        self.credential = self.credentials_provider.get_credentials()

        sign = PreSignedPostSignatureOutPut()
        sign.algorithm = "TOS4-HMAC-SHA256"
        sign.date = date
        sign.credential = self._credential(date)
        sign.origin_policy = _get_post_policy(date, expires, sign.algorithm, sign.credential, bucket, key,
                                              conditions, content_length_range,
                                              self.credential.get_security_token())
        sign.origin_policy = json.dumps(sign.origin_policy)
        sign.policy = base64.b64encode(sign.origin_policy.encode('utf-8')).decode('utf-8')
        sign.signature = self._make_signature(date=date, string_to_sign=sign.policy)

        return sign

    def x_tos_post_sign(self, expires: int, conditions: []):
        if expires is None:
            expires = 60 * 60
        date = datetime.datetime.utcnow().strftime(DATE_FORMAT)
        self.credential = self.credentials_provider.get_credentials()
        params = {}
        params['X-Tos-Algorithm'] = 'TOS4-HMAC-SHA256'
        params['X-Tos-Credential'] = self._credential(date)
        params['X-Tos-Date'] = date
        params['X-Tos-Expires'] = expires
        if self.credential.get_security_token():
            params["X-Tos-Security-Token"] = self.credential.get_security_token()
        params['X-Tos-Policy'] = base64.b64encode(json.dumps(_get_policy(conditions)).encode('utf-8')).decode('utf-8')
        params['X-Tos-Signature'] = self._make_x_tos_policy_signature(date=date, params=params)

        return '&'.join(_param_to_quoted_query(k, v) for k, v in params.items())

    def _make_signature(self, date, req=None, string_to_sign=None):
        if not string_to_sign:
            canonical_request = _canonical_request(req)
            logger.debug("pre-request: canonical_request:\n%s", canonical_request)
            string_to_sign = self._string_to_sign(canonical_request, date)
        logger.debug("pre-request: string_to_sign:\n%s", string_to_sign)
        signature = self._signature(string_to_sign, date)
        logger.debug("pre-request: signature:\n%s", signature)
        return signature

    def _make_x_tos_policy_signature(self, date, params):
        canonical_request = _x_tos_policy_canonical_request(params)
        logger.debug("pre-request: canonical_request:\n%s", canonical_request)
        string_to_sign = self._string_to_sign(canonical_request, date)
        logger.debug("pre-request: string_to_sign:\n%s", string_to_sign)
        signature = self._signature(string_to_sign, date)
        logger.debug("pre-request: signature:\n%s", signature)
        return signature

    def _inject_signature_to_request(self, req, signature, date):
        results = ['TOS4-HMAC-SHA256 Credential=%s' % self._credential(date),
                   'SignedHeaders=%s' % _signed_headers(req.headers), 'Signature=%s' % signature]
        return ', '.join(results)

    def _string_to_sign(self, canonical_request, date):
        sts = ['TOS4-HMAC-SHA256', date, self._credential_scope(date),
               sha256(canonical_request.encode('utf-8')).hexdigest()]
        return '\n'.join(sts)

    def _credential(self, date):
        return "{0}/{1}/{2}/tos/request".format(self.credential.get_access_key_id(), date[0:8], self.region)

    def _credential_scope(self, date):
        return "{0}/{1}/tos/request".format(date[0:8], self.region)

    def _signature(self, string_to_sign, date):
        k_date = _sign(to_bytes(self.credential.get_access_key_secret()), date[0:8])
        k_region = _sign(k_date, self.region)
        k_service = _sign(k_region, 'tos')
        k_signing = _sign(k_service, 'request')
        return _sign(k_signing, string_to_sign, hex=True)

    def copy(self):
        if not isinstance(self.credentials_provider, StaticCredentials):
            return None
        provider = self.credentials_provider.credentials
        return provider.access_key_id, provider.access_key_secret, provider.security_token, self.region


class Auth(AuthBase):
    def __init__(self, access_key_id, access_key_secret, region, sts=None):
        super(Auth, self).__init__(StaticCredentials(access_key_id, access_key_secret, sts), region)


class FederationAuth(AuthBase):
    def __init__(self, credentials_provider: FederationCredentials, region: str):
        super(FederationAuth, self).__init__(credentials_provider, region)



TYPES_MAP = {
    '.ez': 'application/andrew-inset',
    '.anx': 'application/annodex',
    '.atom': 'application/atom+xml',
    '.atomcat': 'application/atomcat+xml',
    '.atomsrv': 'application/atomserv+xml',
    '.lin': 'application/bbolin',
    '.cu': 'application/cu-seeme',
    '.davmount': 'application/davmount+xml',
    '.dcm': 'application/dicom',
    '.tsp': 'application/dsptype',
    '.es': 'application/ecmascript',
    '.otf': 'application/font-sfnt',
    '.ttf': 'application/font-sfnt',
    '.pfr': 'application/font-tdpfr',
    '.woff': 'application/font-woff',
    '.spl': 'application/futuresplash',
    '.gz': 'application/gzip',
    '.hta': 'application/hta',
    '.jar': 'application/java-archive',
    '.ser': 'application/java-serialized-object',
    '.class': 'application/java-vm',
    '.js': 'application/javascript',
    '.json': 'application/json',
    '.m3g': 'application/m3g',
    '.hqx': 'application/mac-binhex40',
    '.cpt': 'application/mac-compactpro',
    '.nb': 'application/mathematica',
    '.nbp': 'application/mathematica',
    '.mbox': 'application/mbox',
    '.mdb': 'application/msaccess',
    '.doc': 'application/msword',
    '.dot': 'application/msword',
    '.mxf': 'application/mxf',
    '.bin': 'application/octet-stream',
    '.oda': 'application/oda',
    '.opf': 'application/oebps-package+xml',
    '.ogx': 'application/ogg',
    '.one': 'application/onenote',
    '.onetoc2': 'application/onenote',
    '.onetmp': 'application/onenote',
    '.onepkg': 'application/onenote',
    '.pdf': 'application/pdf',
    '.pgp': 'application/pgp-encrypted',
    '.key': 'application/pgp-keys',
    '.sig': 'application/pgp-signature',
    '.prf': 'application/pics-rules',
    '.ps': 'application/postscript',
    '.ai': 'application/postscript',
    '.eps': 'application/postscript',
    '.epsi': 'application/postscript',
    '.epsf': 'application/postscript',
    '.eps2': 'application/postscript',
    '.eps3': 'application/postscript',
    '.rar': 'application/rar',
    '.rdf': 'application/rdf+xml',
    '.rtf': 'application/rtf',
    '.stl': 'application/sla',
    '.smi': 'application/smil+xml',
    '.smil': 'application/smil+xml',
    '.apk': 'application/vnd.android.package-archive',
    '.cdy': 'application/vnd.cinderella',
    '.deb': 'application/vnd.debian.binary-package',
    '.ddeb': 'application/vnd.debian.binary-package',
    '.udeb': 'application/vnd.debian.binary-package',
    '.sfd': 'application/vnd.font-fontforge-sfd',
    '.kml': 'application/vnd.google-earth.kml+xml',
    '.kmz': 'application/vnd.google-earth.kmz',
    '.xul': 'application/vnd.mozilla.xul+xml',
    '.xls': 'application/vnd.ms-excel',
    '.xlb': 'application/vnd.ms-excel',
    '.xlt': 'application/vnd.ms-excel',
    '.xlam': 'application/vnd.ms-excel.addin.macroEnabled.12',
    '.xlsb': 'application/vnd.ms-excel.sheet.binary.macroEnabled.12',
    '.xlsm': 'application/vnd.ms-excel.sheet.macroEnabled.12',
    '.xltm': 'application/vnd.ms-excel.template.macroEnabled.12',
    '.eot': 'application/vnd.ms-fontobject',
    '.thmx': 'application/vnd.ms-officetheme',
    '.cat': 'application/vnd.ms-pki.seccat',
    '.ppt': 'application/vnd.ms-powerpoint',
    '.pps': 'application/vnd.ms-powerpoint',
    '.ppam': 'application/vnd.ms-powerpoint.addin.macroEnabled.12',
    '.pptm': 'application/vnd.ms-powerpoint.presentation.macroEnabled.12',
    '.sldm': 'application/vnd.ms-powerpoint.slide.macroEnabled.12',
    '.ppsm': 'application/vnd.ms-powerpoint.slideshow.macroEnabled.12',
    '.potm': 'application/vnd.ms-powerpoint.template.macroEnabled.12',
    '.docm': 'application/vnd.ms-word.document.macroEnabled.12',
    '.dotm': 'application/vnd.ms-word.template.macroEnabled.12',
    '.odc': 'application/vnd.oasis.opendocument.chart',
    '.odb': 'application/vnd.oasis.opendocument.database',
    '.odf': 'application/vnd.oasis.opendocument.formula',
    '.odg': 'application/vnd.oasis.opendocument.graphics',
    '.otg': 'application/vnd.oasis.opendocument.graphics-template',
    '.odi': 'application/vnd.oasis.opendocument.image',
    '.odp': 'application/vnd.oasis.opendocument.presentation',
    '.otp': 'application/vnd.oasis.opendocument.presentation-template',
    '.ods': 'application/vnd.oasis.opendocument.spreadsheet',
    '.ots': 'application/vnd.oasis.opendocument.spreadsheet-template',
    '.odt': 'application/vnd.oasis.opendocument.text',
    '.odm': 'application/vnd.oasis.opendocument.text-master',
    '.ott': 'application/vnd.oasis.opendocument.text-template',
    '.oth': 'application/vnd.oasis.opendocument.text-web',
    '.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    '.sldx': 'application/vnd.openxmlformats-officedocument.presentationml.slide',
    '.ppsx': 'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
    '.potx': 'application/vnd.openxmlformats-officedocument.presentationml.template',
    '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    '.xltx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
    '.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    '.dotx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
    '.cod': 'application/vnd.rim.cod',
    '.mmf': 'application/vnd.smaf',
    '.sdc': 'application/vnd.stardivision.calc',
    '.sds': 'application/vnd.stardivision.chart',
    '.sda': 'application/vnd.stardivision.draw',
    '.sdd': 'application/vnd.stardivision.impress',
    '.sdf': 'application/vnd.stardivision.math',
    '.sdw': 'application/vnd.stardivision.writer',
    '.sgl': 'application/vnd.stardivision.writer-global',
    '.sxc': 'application/vnd.sun.xml.calc',
    '.stc': 'application/vnd.sun.xml.calc.template',
    '.sxd': 'application/vnd.sun.xml.draw',
    '.std': 'application/vnd.sun.xml.draw.template',
    '.sxi': 'application/vnd.sun.xml.impress',
    '.sti': 'application/vnd.sun.xml.impress.template',
    '.sxm': 'application/vnd.sun.xml.math',
    '.sxw': 'application/vnd.sun.xml.writer',
    '.sxg': 'application/vnd.sun.xml.writer.global',
    '.stw': 'application/vnd.sun.xml.writer.template',
    '.sis': 'application/vnd.symbian.install',
    '.cap': 'application/vnd.tcpdump.pcap',
    '.pcap': 'application/vnd.tcpdump.pcap',
    '.vsd': 'application/vnd.visio',
    '.vst': 'application/vnd.visio',
    '.vsw': 'application/vnd.visio',
    '.vss': 'application/vnd.visio',
    '.wbxml': 'application/vnd.wap.wbxml',
    '.wmlc': 'application/vnd.wap.wmlc',
    '.wmlsc': 'application/vnd.wap.wmlscriptc',
    '.wpd': 'application/vnd.wordperfect',
    '.wp5': 'application/vnd.wordperfect5.1',
    '.wasm': 'application/wasm',
    '.wk': 'application/x-123',
    '.7z': 'application/x-7z-compressed',
    '.abw': 'application/x-abiword',
    '.dmg': 'application/x-apple-diskimage',
    '.bcpio': 'application/x-bcpio',
    '.torrent': 'application/x-bittorrent',
    '.cab': 'application/x-cab',
    '.cbr': 'application/x-cbr',
    '.cbz': 'application/x-cbz',
    '.cdf': 'application/x-cdf',
    '.cda': 'application/x-cdf',
    '.vcd': 'application/x-cdlink',
    '.pgn': 'application/x-chess-pgn',
    '.mph': 'application/x-comsol',
    '.cpio': 'application/x-cpio',
    '.csh': 'application/x-csh',
    '.deb': 'application/x-debian-package',
    '.udeb': 'application/x-debian-package',
    '.dcr': 'application/x-director',
    '.dir': 'application/x-director',
    '.dxr': 'application/x-director',
    '.dms': 'application/x-dms',
    '.wad': 'application/x-doom',
    '.dvi': 'application/x-dvi',
    '.pfa': 'application/x-font',
    '.pfb': 'application/x-font',
    '.gsf': 'application/x-font',
    '.pcf': 'application/x-font-pcf',
    '.pcf.Z': 'application/x-font-pcf',
    '.mm': 'application/x-freemind',
    '.spl': 'application/x-futuresplash',
    '.gan': 'application/x-ganttproject',
    '.gnumeric': 'application/x-gnumeric',
    '.sgf': 'application/x-go-sgf',
    '.gcf': 'application/x-graphing-calculator',
    '.gtar': 'application/x-gtar',
    '.tgz': 'application/x-gtar-compressed',
    '.taz': 'application/x-gtar-compressed',
    '.hdf': 'application/x-hdf',
    '.hwp': 'application/x-hwp',
    '.ica': 'application/x-ica',
    '.info': 'application/x-info',
    '.ins': 'application/x-internet-signup',
    '.isp': 'application/x-internet-signup',
    '.iii': 'application/x-iphone',
    '.iso': 'application/x-iso9660-image',
    '.jam': 'application/x-jam',
    '.jnlp': 'application/x-java-jnlp-file',
    '.jmz': 'application/x-jmol',
    '.chrt': 'application/x-kchart',
    '.kil': 'application/x-killustrator',
    '.skp': 'application/x-koan',
    '.skd': 'application/x-koan',
    '.skt': 'application/x-koan',
    '.skm': 'application/x-koan',
    '.kpr': 'application/x-kpresenter',
    '.kpt': 'application/x-kpresenter',
    '.ksp': 'application/x-kspread',
    '.kwd': 'application/x-kword',
    '.kwt': 'application/x-kword',
    '.latex': 'application/x-latex',
    '.lha': 'application/x-lha',
    '.lyx': 'application/x-lyx',
    '.lzh': 'application/x-lzh',
    '.lzx': 'application/x-lzx',
    '.frm': 'application/x-maker',
    '.maker': 'application/x-maker',
    '.frame': 'application/x-maker',
    '.fm': 'application/x-maker',
    '.fb': 'application/x-maker',
    '.book': 'application/x-maker',
    '.fbdoc': 'application/x-maker',
    '.mif': 'application/x-mif',
    '.m3u8': 'application/x-mpegURL',
    '.wmd': 'application/x-ms-wmd',
    '.wmz': 'application/x-ms-wmz',
    '.com': 'application/x-msdos-program',
    '.exe': 'application/x-msdos-program',
    '.bat': 'application/x-msdos-program',
    '.dll': 'application/x-msdos-program',
    '.msi': 'application/x-msi',
    '.nc': 'application/x-netcdf',
    '.pac': 'application/x-ns-proxy-autoconfig',
    '.nwc': 'application/x-nwc',
    '.o': 'application/x-object',
    '.oza': 'application/x-oz-application',
    '.p7r': 'application/x-pkcs7-certreqresp',
    '.crl': 'application/x-pkcs7-crl',
    '.pyc': 'application/x-python-code',
    '.pyo': 'application/x-python-code',
    '.qgs': 'application/x-qgis',
    '.shp': 'application/x-qgis',
    '.shx': 'application/x-qgis',
    '.qtl': 'application/x-quicktimeplayer',
    '.rdp': 'application/x-rdp',
    '.rpm': 'application/x-redhat-package-manager',
    '.rss': 'application/x-rss+xml',
    '.rb': 'application/x-ruby',
    '.sci': 'application/x-scilab',
    '.sce': 'application/x-scilab',
    '.xcos': 'application/x-scilab-xcos',
    '.sh': 'application/x-sh',
    '.shar': 'application/x-shar',
    '.swf': 'application/x-shockwave-flash',
    '.swfl': 'application/x-shockwave-flash',
    '.scr': 'application/x-silverlight',
    '.sql': 'application/x-sql',
    '.sit': 'application/x-stuffit',
    '.sitx': 'application/x-stuffit',
    '.sv4cpio': 'application/x-sv4cpio',
    '.sv4crc': 'application/x-sv4crc',
    '.tar': 'application/x-tar',
    '.tcl': 'application/x-tcl',
    '.gf': 'application/x-tex-gf',
    '.pk': 'application/x-tex-pk',
    '.texinfo': 'application/x-texinfo',
    '.texi': 'application/x-texinfo',
    '.~': 'application/x-trash',
    '.%': 'application/x-trash',
    '.bak': 'application/x-trash',
    '.old': 'application/x-trash',
    '.sik': 'application/x-trash',
    '.t': 'application/x-troff',
    '.tr': 'application/x-troff',
    '.roff': 'application/x-troff',
    '.man': 'application/x-troff-man',
    '.me': 'application/x-troff-me',
    '.ms': 'application/x-troff-ms',
    '.ustar': 'application/x-ustar',
    '.src': 'application/x-wais-source',
    '.wz': 'application/x-wingz',
    '.crt': 'application/x-x509-ca-cert',
    '.xcf': 'application/x-xcf',
    '.fig': 'application/x-xfig',
    '.xpi': 'application/x-xpinstall',
    '.xz': 'application/x-xz',
    '.xhtml': 'application/xhtml+xml',
    '.xht': 'application/xhtml+xml',
    '.xml': 'application/xml',
    '.xsd': 'application/xml',
    '.xsl': 'application/xslt+xml',
    '.xslt': 'application/xslt+xml',
    '.xspf': 'application/xspf+xml',
    '.zip': 'application/zip',
    '.amr': 'audio/amr',
    '.awb': 'audio/amr-wb',
    '.axa': 'audio/annodex',
    '.au': 'audio/basic',
    '.snd': 'audio/basic',
    '.csd': 'audio/csound',
    '.orc': 'audio/csound',
    '.sco': 'audio/csound',
    '.flac': 'audio/flac',
    '.mid': 'audio/midi',
    '.midi': 'audio/midi',
    '.kar': 'audio/midi',
    '.mpga': 'audio/mpeg',
    '.mpega': 'audio/mpeg',
    '.mp2': 'audio/mpeg',
    '.mp3': 'audio/mpeg',
    '.m4a': 'audio/mpeg',
    '.m3u': 'audio/mpegurl',
    '.oga': 'audio/ogg',
    '.ogg': 'audio/ogg',
    '.opus': 'audio/ogg',
    '.spx': 'audio/ogg',
    '.sid': 'audio/prs.sid',
    '.aif': 'audio/x-aiff',
    '.aiff': 'audio/x-aiff',
    '.aifc': 'audio/x-aiff',
    '.gsm': 'audio/x-gsm',
    '.m3u': 'audio/x-mpegurl',
    '.wax': 'audio/x-ms-wax',
    '.wma': 'audio/x-ms-wma',
    '.ra': 'audio/x-pn-realaudio',
    '.rm': 'audio/x-pn-realaudio',
    '.ram': 'audio/x-pn-realaudio',
    '.ra': 'audio/x-realaudio',
    '.pls': 'audio/x-scpls',
    '.sd2': 'audio/x-sd2',
    '.wav': 'audio/x-wav',
    '.alc': 'chemical/x-alchemy',
    '.cac': 'chemical/x-cache',
    '.cache': 'chemical/x-cache',
    '.csf': 'chemical/x-cache-csf',
    '.cbin': 'chemical/x-cactvs-binary',
    '.cascii': 'chemical/x-cactvs-binary',
    '.ctab': 'chemical/x-cactvs-binary',
    '.cdx': 'chemical/x-cdx',
    '.cer': 'chemical/x-cerius',
    '.c3d': 'chemical/x-chem3d',
    '.chm': 'chemical/x-chemdraw',
    '.cif': 'chemical/x-cif',
    '.cmdf': 'chemical/x-cmdf',
    '.cml': 'chemical/x-cml',
    '.cpa': 'chemical/x-compass',
    '.bsd': 'chemical/x-crossfire',
    '.csml': 'chemical/x-csml',
    '.csm': 'chemical/x-csml',
    '.ctx': 'chemical/x-ctx',
    '.cxf': 'chemical/x-cxf',
    '.cef': 'chemical/x-cxf',
    '.emb': 'chemical/x-embl-dl-nucleotide',
    '.embl': 'chemical/x-embl-dl-nucleotide',
    '.spc': 'chemical/x-galactic-spc',
    '.inp': 'chemical/x-gamess-input',
    '.gam': 'chemical/x-gamess-input',
    '.gamin': 'chemical/x-gamess-input',
    '.fch': 'chemical/x-gaussian-checkpoint',
    '.fchk': 'chemical/x-gaussian-checkpoint',
    '.cub': 'chemical/x-gaussian-cube',
    '.gau': 'chemical/x-gaussian-input',
    '.gjc': 'chemical/x-gaussian-input',
    '.gjf': 'chemical/x-gaussian-input',
    '.gal': 'chemical/x-gaussian-log',
    '.gcg': 'chemical/x-gcg8-sequence',
    '.gen': 'chemical/x-genbank',
    '.hin': 'chemical/x-hin',
    '.istr': 'chemical/x-isostar',
    '.ist': 'chemical/x-isostar',
    '.jdx': 'chemical/x-jcamp-dx',
    '.dx': 'chemical/x-jcamp-dx',
    '.kin': 'chemical/x-kinemage',
    '.mcm': 'chemical/x-macmolecule',
    '.mmd': 'chemical/x-macromodel-input',
    '.mmod': 'chemical/x-macromodel-input',
    '.mol': 'chemical/x-mdl-molfile',
    '.rd': 'chemical/x-mdl-rdfile',
    '.rxn': 'chemical/x-mdl-rxnfile',
    '.sd': 'chemical/x-mdl-sdfile',
    '.sdf': 'chemical/x-mdl-sdfile',
    '.tgf': 'chemical/x-mdl-tgf',
    '.mcif': 'chemical/x-mmcif',
    '.mol2': 'chemical/x-mol2',
    '.b': 'chemical/x-molconn-Z',
    '.gpt': 'chemical/x-mopac-graph',
    '.mop': 'chemical/x-mopac-input',
    '.mopcrt': 'chemical/x-mopac-input',
    '.mpc': 'chemical/x-mopac-input',
    '.zmt': 'chemical/x-mopac-input',
    '.moo': 'chemical/x-mopac-out',
    '.mvb': 'chemical/x-mopac-vib',
    '.asn': 'chemical/x-ncbi-asn1',
    '.prt': 'chemical/x-ncbi-asn1-ascii',
    '.ent': 'chemical/x-ncbi-asn1-ascii',
    '.val': 'chemical/x-ncbi-asn1-binary',
    '.aso': 'chemical/x-ncbi-asn1-binary',
    '.asn': 'chemical/x-ncbi-asn1-spec',
    '.pdb': 'chemical/x-pdb',
    '.ent': 'chemical/x-pdb',
    '.ros': 'chemical/x-rosdal',
    '.sw': 'chemical/x-swissprot',
    '.vms': 'chemical/x-vamas-iso14976',
    '.vmd': 'chemical/x-vmd',
    '.xtel': 'chemical/x-xtel',
    '.xyz': 'chemical/x-xyz',
    '.gif': 'image/gif',
    '.ief': 'image/ief',
    '.jp2': 'image/jp2',
    '.jpg2': 'image/jp2',
    '.jpeg': 'image/jpeg',
    '.jpg': 'image/jpeg',
    '.jpe': 'image/jpeg',
    '.jpm': 'image/jpm',
    '.jpx': 'image/jpx',
    '.jpf': 'image/jpx',
    '.pcx': 'image/pcx',
    '.png': 'image/png',
    '.svg': 'image/svg+xml',
    '.svgz': 'image/svg+xml',
    '.tiff': 'image/tiff',
    '.tif': 'image/tiff',
    '.djvu': 'image/vnd.djvu',
    '.djv': 'image/vnd.djvu',
    '.ico': 'image/vnd.microsoft.icon',
    '.wbmp': 'image/vnd.wap.wbmp',
    '.cr2': 'image/x-canon-cr2',
    '.crw': 'image/x-canon-crw',
    '.ras': 'image/x-cmu-raster',
    '.cdr': 'image/x-coreldraw',
    '.pat': 'image/x-coreldrawpattern',
    '.cdt': 'image/x-coreldrawtemplate',
    '.cpt': 'image/x-corelphotopaint',
    '.erf': 'image/x-epson-erf',
    '.art': 'image/x-jg',
    '.jng': 'image/x-jng',
    '.bmp': 'image/x-ms-bmp',
    '.nef': 'image/x-nikon-nef',
    '.orf': 'image/x-olympus-orf',
    '.psd': 'image/x-photoshop',
    '.pnm': 'image/x-portable-anymap',
    '.pbm': 'image/x-portable-bitmap',
    '.pgm': 'image/x-portable-graymap',
    '.ppm': 'image/x-portable-pixmap',
    '.rgb': 'image/x-rgb',
    '.xbm': 'image/x-xbitmap',
    '.xpm': 'image/x-xpixmap',
    '.xwd': 'image/x-xwindowdump',
    '.eml': 'message/rfc822',
    '.igs': 'model/iges',
    '.iges': 'model/iges',
    '.msh': 'model/mesh',
    '.mesh': 'model/mesh',
    '.silo': 'model/mesh',
    '.wrl': 'model/vrml',
    '.vrml': 'model/vrml',
    '.x3db': 'model/x3d+binary',
    '.x3dv': 'model/x3d+vrml',
    '.x3d': 'model/x3d+xml',
    '.appcache': 'text/cache-manifest',
    '.ics': 'text/calendar',
    '.icz': 'text/calendar',
    '.css': 'text/css',
    '.csv': 'text/csv',
    '.323': 'text/h323',
    '.html': 'text/html',
    '.htm': 'text/html',
    '.shtml': 'text/html',
    '.uls': 'text/iuls',
    '.mml': 'text/mathml',
    '.asc': 'text/plain',
    '.txt': 'text/plain',
    '.text': 'text/plain',
    '.pot': 'text/plain',
    '.brf': 'text/plain',
    '.srt': 'text/plain',
    '.rtx': 'text/richtext',
    '.sct': 'text/scriptlet',
    '.wsc': 'text/scriptlet',
    '.tsv': 'text/tab-separated-values',
    '.tm': 'text/texmacs',
    '.ttl': 'text/turtle',
    '.vcf': 'text/vcard',
    '.vcard': 'text/vcard',
    '.jad': 'text/vnd.sun.j2me.app-descriptor',
    '.wml': 'text/vnd.wap.wml',
    '.wmls': 'text/vnd.wap.wmlscript',
    '.bib': 'text/x-bibtex',
    '.boo': 'text/x-boo',
    '.h++': 'text/x-c++hdr',
    '.hpp': 'text/x-c++hdr',
    '.hxx': 'text/x-c++hdr',
    '.hh': 'text/x-c++hdr',
    '.c++': 'text/x-c++src',
    '.cpp': 'text/x-c++src',
    '.cxx': 'text/x-c++src',
    '.cc': 'text/x-c++src',
    '.h': 'text/x-chdr',
    '.htc': 'text/x-component',
    '.csh': 'text/x-csh',
    '.c': 'text/x-csrc',
    '.diff': 'text/x-diff',
    '.patch': 'text/x-diff',
    '.d': 'text/x-dsrc',
    '.hs': 'text/x-haskell',
    '.java': 'text/x-java',
    '.ly': 'text/x-lilypond',
    '.lhs': 'text/x-literate-haskell',
    '.moc': 'text/x-moc',
    '.p': 'text/x-pascal',
    '.pas': 'text/x-pascal',
    '.gcd': 'text/x-pcs-gcd',
    '.pl': 'text/x-perl',
    '.pm': 'text/x-perl',
    '.py': 'text/x-python',
    '.scala': 'text/x-scala',
    '.etx': 'text/x-setext',
    '.sfv': 'text/x-sfv',
    '.sh': 'text/x-sh',
    '.tcl': 'text/x-tcl',
    '.tk': 'text/x-tcl',
    '.tex': 'text/x-tex',
    '.ltx': 'text/x-tex',
    '.sty': 'text/x-tex',
    '.cls': 'text/x-tex',
    '.vcs': 'text/x-vcalendar',
    '.3gp': 'video/3gpp',
    '.ts': 'video/MP2T',
    '.axv': 'video/annodex',
    '.dl': 'video/dl',
    '.dif': 'video/dv',
    '.dv': 'video/dv',
    '.fli': 'video/fli',
    '.gl': 'video/gl',
    '.mp4': 'video/mp4',
    '.mpeg': 'video/mpeg',
    '.mpg': 'video/mpeg',
    '.mpe': 'video/mpeg',
    '.ogv': 'video/ogg',
    '.qt': 'video/quicktime',
    '.mov': 'video/quicktime',
    '.mxu': 'video/vnd.mpegurl',
    '.webm': 'video/webm',
    '.flv': 'video/x-flv',
    '.lsf': 'video/x-la-asf',
    '.lsx': 'video/x-la-asf',
    '.mpv': 'video/x-matroska',
    '.mkv': 'video/x-matroska',
    '.mng': 'video/x-mng',
    '.asf': 'video/x-ms-asf',
    '.asx': 'video/x-ms-asf',
    '.wm': 'video/x-ms-wm',
    '.wmv': 'video/x-ms-wmv',
    '.wmx': 'video/x-ms-wmx',
    '.wvx': 'video/x-ms-wvx',
    '.avi': 'video/x-msvideo',
    '.movie': 'video/x-sgi-movie',
    '.ice': 'x-conference/x-cooltalk',
    '.sisx': 'x-epoc/x-sisx-app',
    '.vrm': 'x-world/x-vrml',
    '.vrml': 'x-world/x-vrml',
    '.wrl': 'x-world/x-vrml'
}



from .consts import LAST_MODIFY_TIME_DATE_FORMAT
from .models2 import Owner, RedirectAllRequestsTo, IndexDocument, ErrorDocument, RoutingRules, CustomDomainRule, \
    RealTimeLogConfiguration
from .utils import check_enum_type


def to_complete_multipart_upload_request(parts: list):
    dic = {}
    p = []
    for part in parts:
        p.append({'PartNumber': part.part_number, 'ETag': part.etag})
    dic['Parts'] = p
    return dic


def to_delete_multi_objects_request(objects: [], quiet: bool):
    data = {}
    if objects:
        obs = []
        for o in objects:
            obs.append({'Key': o.key, 'VersionId': o.version_id})
        data['Objects'] = obs
    data['Quiet'] = quiet
    return data


def to_put_acl_request(owner: Owner, grants: []):
    data = {}
    if owner:
        data['Owner'] = {"ID": owner.id, "DisplayName": owner.display_name}
    if grants:
        convertor_grant = []
        for grant in grants:
            m = {}
            grantee = grant.grantee
            check_enum_type(grantee=grantee.type, canned=grantee.canned, permission=grant.permission)
            m['Grantee'] = {'ID': grantee.id, 'Type': grantee.type.value, 'DisplayName': grantee.display_name}
            if grantee.canned:
                m['Grantee']['Canned'] = grantee.canned.value

            m['Permission'] = grant.permission.value
            convertor_grant.append(m)
        data['Grants'] = convertor_grant

    return data


def to_put_bucket_cors_request(cors_rules: []):
    data = {}
    arr = []
    for cors_rule in cors_rules:
        info = {}
        if cors_rule.allowed_origins:
            info['AllowedOrigins'] = cors_rule.allowed_origins
        if cors_rule.allowed_methods:
            info['AllowedMethods'] = cors_rule.allowed_methods
        if cors_rule.allowed_headers:
            info['AllowedHeaders'] = cors_rule.allowed_headers
        if cors_rule.expose_headers:
            info['ExposeHeaders'] = cors_rule.expose_headers
        if cors_rule.max_age_seconds:
            info['MaxAgeSeconds'] = cors_rule.max_age_seconds
        arr.append(info)

    data['CORSRules'] = arr

    return data


def to_put_bucket_mirror_back(rules: []):
    data = {}
    arr = []
    for rule in rules:
        info = {}
        if rule.id:
            info['ID'] = rule.id
        if rule.condition:
            info['Condition'] = {}
            if rule.condition.http_code:
                info['Condition']['HttpCode'] = rule.condition.http_code
        if rule.redirect:
            info['Redirect'] = {}
            if rule.redirect.redirect_type:
                info['Redirect']['RedirectType'] = rule.redirect.redirect_type.value
            if rule.redirect.fetch_source_on_redirect is not None:
                info['Redirect']['FetchSourceOnRedirect'] = rule.redirect.fetch_source_on_redirect
            if rule.redirect.public_source:
                info['Redirect']['PublicSource'] = {}
                if rule.redirect.public_source.source_endpoint:
                    info['Redirect']['PublicSource']['SourceEndpoint'] = {}
                    if rule.redirect.public_source.source_endpoint.primary:
                        info['Redirect']['PublicSource']['SourceEndpoint'][
                            'Primary'] = rule.redirect.public_source.source_endpoint.primary
                    if rule.redirect.public_source.source_endpoint.follower:
                        info['Redirect']['PublicSource']['SourceEndpoint'][
                            'Follower'] = rule.redirect.public_source.source_endpoint.follower
            if rule.redirect.pass_query is not None:
                info['Redirect']['PassQuery'] = rule.redirect.pass_query
            if rule.redirect.follow_redirect is not None:
                info['Redirect']['FollowRedirect'] = rule.redirect.follow_redirect
            if rule.redirect.mirror_header:
                info['Redirect']['MirrorHeader'] = {}
                if rule.redirect.mirror_header.pass_all:
                    info['Redirect']['MirrorHeader']['PassAll'] = rule.redirect.mirror_header.pass_all
                if rule.redirect.mirror_header.pass_headers:
                    info['Redirect']['MirrorHeader']['Pass'] = rule.redirect.mirror_header.pass_headers
                if rule.redirect.mirror_header.remove:
                    info['Redirect']['MirrorHeader']['Remove'] = rule.redirect.mirror_header.remove
        arr.append(info)

    data['Rules'] = arr

    return data


def to_put_bucket_lifecycle(rules: []):
    data = {}
    arr = []
    for rule in rules:
        info = {}

        if rule.id:
            info['ID'] = rule.id

        if rule.prefix:
            info['Prefix'] = rule.prefix

        if rule.status:
            info['Status'] = rule.status.value

        if rule.expiration:
            info['Expiration'] = {}
            if rule.expiration.days:
                info['Expiration']['Days'] = rule.expiration.days
            if rule.expiration.date:
                info['Expiration']['Date'] = rule.expiration.date.strftime(LAST_MODIFY_TIME_DATE_FORMAT)

        if rule.no_current_version_expiration:
            info['NoncurrentVersionExpiration'] = {
                'NoncurrentDays': rule.no_current_version_expiration.no_current_days
            }

        if rule.abort_in_complete_multipart_upload and rule.abort_in_complete_multipart_upload.days_after_init:
            info['AbortIncompleteMultipartUpload'] = {
                'DaysAfterInitiation': rule.abort_in_complete_multipart_upload.days_after_init
            }

        if rule.tags:
            tag_arr = []
            for t in rule.tags:
                tag_arr.append({
                    'Key': t.key,
                    'Value': t.value
                })
            info['Tags'] = tag_arr

        if rule.transitions:
            trans_arr = []
            for t in rule.transitions:
                transition_info = {}
                if t.days:
                    transition_info['Days'] = t.days
                if t.date:
                    transition_info['Date'] = t.date.strftime(LAST_MODIFY_TIME_DATE_FORMAT)
                if t.storage_class:
                    transition_info['StorageClass'] = t.storage_class.value
                trans_arr.append(transition_info)
            info['Transitions'] = trans_arr

        if rule.non_current_version_transitions:
            current_version_transition_arr = []
            for tr in rule.non_current_version_transitions:
                non_current_version_transition_info = {}
                if tr.non_current_days:
                    non_current_version_transition_info['NoncurrentDays'] = tr.non_current_days

                if tr.storage_class:
                    non_current_version_transition_info['StorageClass'] = tr.storage_class.value

                current_version_transition_arr.append(non_current_version_transition_info)

            info['NoncurrentVersionTransitions'] = current_version_transition_arr

        arr.append(info)

    data['Rules'] = arr

    return data


def to_put_object_tagging(tags: []):
    info = []
    for tag in tags:
        info.append({
            'Key': tag.key,
            'Value': tag.value
        })
    data = {'TagSet': {
        'Tags': info
    }}
    return data


def to_fetch_object(url: str, object: str = None, ignore_same_key=None, content_md5=None):
    info = {'URL': url}
    if object:
        info['Object'] = object
    if ignore_same_key:
        info['IgnoreSameKey'] = ignore_same_key
    if content_md5:
        info['ContentMD5'] = content_md5

    return info


def to_put_replication(role: str, rules: []):
    info = {}
    if role:
        info['Role'] = role
    r = []
    for rule in rules:
        data = {}
        if rule.id:
            data['ID'] = rule.id
        if rule.status:
            data['Status'] = rule.status.value
        if rule.prefix_set:
            data['PrefixSet'] = rule.prefix_set
        if rule.destination:
            data['Destination'] = {}
            if rule.destination.bucket:
                data['Destination']['Bucket'] = rule.destination.bucket
            if rule.destination.location:
                data['Destination']['Location'] = rule.destination.location
            if rule.destination.storage_class:
                data['Destination']['StorageClass'] = rule.destination.storage_class.value
            if rule.destination.storage_class_inherit_directive:
                data['Destination'][
                    'StorageClassInheritDirective'] = rule.destination.storage_class_inherit_directive.value
        if rule.historical_object_replication:
            data['HistoricalObjectReplication'] = rule.historical_object_replication.value
        if rule.progress:
            data['Progress'] = {}
            if rule.progress.historical_object:
                data['Progress']['HistoricalObject'] = rule.progress.historical_object
            if rule.progress.new_object:
                data['Progress']['NewObject'] = rule.progress.new_object

        r.append(data)
    info['Rules'] = r
    return info


def to_put_bucket_website(redirect_all_requests_to: RedirectAllRequestsTo,
                          index_document: IndexDocument, error_document: ErrorDocument, routing_rules: RoutingRules):
    info = {}
    if redirect_all_requests_to:
        info['RedirectAllRequestsTo'] = {}
        if redirect_all_requests_to.host_name:
            info['RedirectAllRequestsTo']['HostName'] = redirect_all_requests_to.host_name
        if redirect_all_requests_to.protocol:
            info['RedirectAllRequestsTo']['Protocol'] = redirect_all_requests_to.protocol
    if index_document:
        info['IndexDocument'] = {}
        if index_document.suffix:
            info['IndexDocument']['Suffix'] = index_document.suffix
        if index_document.forbidden_sub_dir:
            info['IndexDocument']['ForbiddenSubDir'] = index_document.forbidden_sub_dir

    if error_document:
        info['ErrorDocument'] = {}
        if error_document.key:
            info['ErrorDocument']['Key'] = error_document.key

    if routing_rules:
        info['RoutingRules'] = []
        if routing_rules.rules and len(routing_rules.rules) > 0:
            for rule in routing_rules.rules:
                rule_mp = {}
                if rule.condition:
                    rule_mp['Condition'] = {}
                    if rule.condition.key_prefix_equals:
                        rule_mp['Condition']['KeyPrefixEquals'] = rule.condition.key_prefix_equals
                    if rule.condition.http_error_code_returned_equals:
                        rule_mp['Condition'][
                            'HttpErrorCodeReturnedEquals'] = rule.condition.http_error_code_returned_equals
                if rule.redirect:
                    rule_mp['Redirect'] = {}
                    if rule.redirect.host_name:
                        rule_mp['Redirect']['HostName'] = rule.redirect.host_name
                    if rule.redirect.http_redirect_code:
                        rule_mp['Redirect']['HttpRedirectCode'] = rule.redirect.http_redirect_code
                    if rule.redirect.protocol:
                        rule_mp['Redirect']['Protocol'] = rule.redirect.protocol.value
                    if rule.redirect.replace_key_prefix_with:
                        rule_mp['Redirect']['ReplaceKeyPrefixWith'] = rule.redirect.replace_key_prefix_with
                    if rule.redirect.replace_key_with:
                        rule_mp['Redirect']['ReplaceKeyWith'] = rule.redirect.replace_key_with

                info['RoutingRules'].append(rule_mp)
    return info


def to_put_bucket_notification(cloudFunctionConfigurations: []):
    info = {}
    if cloudFunctionConfigurations:
        info['CloudFunctionConfigurations'] = []
        for cloudFunctionConfiguration in cloudFunctionConfigurations:
            config = {}
            if cloudFunctionConfiguration.events:
                config['Events'] = cloudFunctionConfiguration.events
            if cloudFunctionConfiguration.id:
                config['RuleId'] = cloudFunctionConfiguration.id
            if cloudFunctionConfiguration.cloud_function:
                config['CloudFunction'] = cloudFunctionConfiguration.cloud_function
            if cloudFunctionConfiguration.filter:
                filter_mp = {}
                if cloudFunctionConfiguration.filter.key:
                    filter_mp['TOSKey'] = {}
                    if cloudFunctionConfiguration.filter.key.rules and len(
                            cloudFunctionConfiguration.filter.key.rules) >= 1:
                        filter_mp['TOSKey']['FilterRules'] = []
                        for rule in cloudFunctionConfiguration.filter.key.rules:
                            rule_mp = {}
                            if rule.name:
                                rule_mp['Name'] = rule.name
                            if rule.value:
                                rule_mp['Value'] = rule.value
                            filter_mp['TOSKey']['FilterRules'].append(rule_mp)

                config['Filter'] = filter_mp
            info['CloudFunctionConfigurations'].append(config)

    return info


def to_put_custom_domain(custom_domain_rule: CustomDomainRule):
    info = {}
    if custom_domain_rule:
        info['CustomDomainRule'] = {}
        if custom_domain_rule.domain:
            info['CustomDomainRule']['Domain'] = custom_domain_rule.domain
        if custom_domain_rule.cname:
            info['CustomDomainRule']['Cname'] = custom_domain_rule.cname
        if custom_domain_rule.cert_id:
            info['CustomDomainRule']['CertId'] = custom_domain_rule.cert_id
        if custom_domain_rule.cert_status:
            info['CustomDomainRule']['CertStatus'] = custom_domain_rule.cert_status.value
        if custom_domain_rule.forbidden:
            info['CustomDomainRule']['Forbidden'] = custom_domain_rule.forbidden
        if custom_domain_rule.forbidden_reason:
            info['CustomDomainRule']['ForbiddenReason'] = custom_domain_rule.forbidden_reason

    return info


def to_put_bucket_real_time_log(configuation: RealTimeLogConfiguration):
    info = {}
    if configuation:
        info['RealTimeLogConfiguration'] = {}
        if configuation.role:
            info['RealTimeLogConfiguration']['Role'] = configuation.role
        if configuation.configuration:
            info['RealTimeLogConfiguration']['AccessLogConfiguration'] = {}
            if configuation.configuration.use_service_topic:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'UseServiceTopic'] = configuation.configuration.use_service_topic
            if configuation.configuration.tls_topic_id:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'TLSTopicID'] = configuation.configuration.tls_topic_id
            if configuation.configuration.tls_project_id:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'TLSProjectID'] = configuation.configuration.tls_project_id
    return info



import concurrent.futures
import hashlib
import json
import logging
import os
import threading
import urllib.parse
from concurrent.futures import as_completed
from typing import Dict

from .enum import DownloadEventType, UploadEventType, CopyEventType
from .exceptions import (CancelNotWithAbortError, CancelWithAbortError,
                         TosClientError, TosServerError, TosError, TaskCompleteMultipartError, RenameFileError)
from .models2 import PartInfo, UploadedPart, _PartToDo, CopyPartInfo
from .utils import (SizeAdapter, to_unicode, MergeProcess, to_bytes,
                    cal_crc_from_download_parts, check_crc, rename_file, copy_and_verify_length, makedir_p,
                    cal_crc_from_upload_parts)

logger = logging.getLogger(__name__)


class CheckPointStore(object):
    """
    checkpoint 相关信息维护类
    创建CheckPointStore对象时，指定目录下创建存储 checkpoint信息的文件目录
    支持 获取指定对象 key 的 checkpoint json 信息
    支持 在指定目录下创建 对象checkpoint 信息
    支持 删除对象的 checkpoint 信息
    """

    def __init__(self, dir, file_name, use_type: str):

        self.dir = dir
        self.file_name = file_name
        self.suffix = use_type

        if os.path.isdir(self.dir):
            return

        makedir_p(self.dir)

    def get(self, bucket, key, src_bucket=None, src_key=None, version_id=None):
        pathname = self.path(bucket, key, src_bucket=src_bucket, src_key=src_key, versionId=version_id)

        if not os.path.exists(pathname):
            return None

        try:
            with open(to_unicode(pathname), 'r') as f:
                content = json.load(f)
        except ValueError:
            os.remove(pathname)
            return None
        else:
            return content

    def put(self, bucket, key, value, src_bucket=None, src_key=None, version_id=None):
        pathname = self.path(bucket, key, src_bucket, src_key, version_id)

        with open(to_unicode(pathname), 'w') as f:
            json.dump(value, f)

    def delete(self, bucket, key, src_bucket=None, src_key=None, version_id=None):
        pathname = self.path(bucket, key, src_bucket=src_bucket, src_key=src_key, versionId=version_id)
        try:
            os.remove(pathname)
        except Exception:
            return

    def path(self, bucket, key, src_bucket=None, src_key=None, versionId=None):
        encode_str = ''
        if src_bucket and src_key:
            encode_str += src_bucket + '.' + src_key
            if versionId:
                encode_str += '.' + versionId
            encode_str += '.' + bucket + '.' + key
        else:
            encode_str += bucket + '.' + key
            if versionId:
                encode_str += '.' + versionId
        encoding_data = urllib.parse.quote(hashlib.md5(to_bytes(self.file_name + encode_str)).digest(), safe='')

        name = "{0}.{1}".format(encoding_data, self.suffix)
        return os.path.join(self.dir, name)


def _cover_to_uploaded_parts(parts: []) -> list:
    """
    将_PartToProcess 转化为 UploadedPart
    """
    l = []
    for p in parts:
        uploaded_part = UploadedPart(p.part_number, p.etag)
        l.append(uploaded_part)

    return l


class CancelHook(object):
    def __init__(self):
        self._is_cal = False
        self._is_abort = None
        self._lock = threading.Lock()

    def cancel(self, is_abort: bool):
        with self._lock:
            self._is_cal = True
            self._is_abort = is_abort

    def _is_abort_func(self):
        if self._is_cal and self._is_abort:
            return True

    def _is_not_abort_func(self):
        if self._is_cal and not self._is_abort:
            return True

    def is_cancel(self):
        with self._lock:
            if self._is_abort_func():
                raise CancelWithAbortError('user cancel upload file task with abort')
            if self._is_not_abort_func():
                raise CancelNotWithAbortError('user cancel upload file task not with abort')


class BreakpointBase(object):
    def __init__(self, client, bucket, key, store: CheckPointStore, task_num,
                 parts_to_do, record: Dict,
                 size, rate_limiter, cancel_hook,
                 datatransfer_listener, event_listener):
        self.client = client
        self.bucket = bucket
        self.key = key
        self.task_num = task_num
        self.parts_to_do = parts_to_do
        self.record = record
        self.size = size
        self.cancel_hook = cancel_hook
        self.datatransfer_listener = datatransfer_listener
        self.event_listener = event_listener
        self.rate_limiter = rate_limiter
        self.need_bytes = 0
        for part in parts_to_do:
            self.need_bytes += part.size

        if self.datatransfer_listener:
            self.datatransfer_listener = MergeProcess(self.datatransfer_listener, size,
                                                      len(self.parts_to_do), size - self.need_bytes)
        # 下列变量加锁操作
        self.lock = threading.Lock()
        self.record = record
        self.store = store

        # 不同实现
        self.finished_parts = []

    def execute(self, **kwargs):
        try:
            self._cover_to_finished_parts()
            self._pre_task()
            q = TaskExecutor(self.task_num, self._do_task, self.cancel_hook)
            for part in self.parts_to_do:
                q.submit(part)

            q.run()
            # 执行任务
            result = self._last_task(**kwargs)
            self._delete_checkpoint()
            self._callback_success()
            return result

        except TaskCompleteMultipartError as e:
            self._callback_fail(e)
            raise TosClientError(e.message, e)
        except CancelWithAbortError as e:
            self._callback_abort(e)
            self._abort_task()
            self._delete_checkpoint()
            raise TosClientError('the task is canceled')
        except CancelNotWithAbortError as e:
            self._callback_abort(e)
            raise TosClientError('the task is canceled')
        except RenameFileError as e:
            self._callback_rename_fail(e)
            raise TosClientError(e.message, e)
        except TosError as e:
            raise e
        except Exception as e:
            raise TosClientError('unknown err', e)

    def _cover_to_finished_parts(self):
        pass

    def _delete_checkpoint(self):
        with self.lock:
            self.store.delete(self.bucket, self.key)

    def _pre_task(self):
        pass

    def _do_task(self, part):
        pass

    def _last_task(self, **args):
        pass

    def _abort_task(self, *args):
        pass

    def _callback_success(self):
        pass

    def _callback_fail(self, e):
        pass

    def _callback_part_fail(self, e, part):
        pass

    def _callback_part_success(self, part_info):
        pass

    def _callback_rename_fail(self, e):
        pass

    def _callback_abort(self, e):
        pass

    def _finish_part(self, part_info):
        pass


class _BreakpointUploader(BreakpointBase):
    def __init__(self, client, bucket, key, store: CheckPointStore, task_num,
                 parts_to_update, upload_id, record: Dict,
                 size, ssec_algorithm, ssec_key, ssec_key_md5, rate_limiter, cancel_hook,
                 datatransfer_listener, upload_event_listener,
                 file_path):

        super(_BreakpointUploader, self).__init__(client=client, bucket=bucket, key=key, store=store, task_num=task_num,
                                                  parts_to_do=parts_to_update, record=record, size=size,
                                                  rate_limiter=rate_limiter, cancel_hook=cancel_hook,
                                                  datatransfer_listener=datatransfer_listener,
                                                  event_listener=upload_event_listener)
        self.filename = file_path
        self.task_num = task_num
        self.upload_id = upload_id
        self.cancel_hook = cancel_hook
        self.ssec_algorithm = ssec_algorithm
        self.ssec_key = ssec_key
        self.ssec_key_md5 = ssec_key_md5

    def _cover_to_finished_parts(self):
        for p in self.record["parts_info"]:
            self.finished_parts.append(
                PartInfo(part_number=p['part_number'], part_size=p['part_size'], offset=p['offset'], etag=p['etag'],
                         hash_crc64_ecma=p['hash_crc64ecma'], is_completed=True))

    def _do_task(self, part):
        with open(to_unicode(self.filename), 'rb') as f:
            f.seek(part.start, os.SEEK_SET)
            try:
                result = self.client.upload_part(bucket=self.bucket, key=self.key, upload_id=self.upload_id,
                                                 part_number=part.part_number,
                                                 content=SizeAdapter(f, part.size, init_offset=part.start,
                                                                     can_reset=True),
                                                 data_transfer_listener=self.datatransfer_listener,
                                                 rate_limiter=self.rate_limiter,
                                                 ssec_algorithm=self.ssec_algorithm,
                                                 ssec_key=self.ssec_key,
                                                 ssec_key_md5=self.ssec_key_md5)
            except Exception as e:
                self._callback_part_fail(e, part)
                raise e

            self._finish_part(
                PartInfo(part_number=part.part_number, part_size=part.size, offset=part.start, etag=result.etag,
                         hash_crc64_ecma=result.hash_crc64_ecma, is_completed=True))

    def _abort_task(self):
        self.client.abort_multipart_upload(self.bucket, self.key, self.upload_id)

    def _delete_checkpoint(self):
        self.store.delete(self.bucket, self.key)

    def _last_task(self):
        try:
            parts = _cover_to_uploaded_parts(self.finished_parts)
            result = self.client.complete_multipart_upload(self.bucket, self.key, self.upload_id, parts=parts)
            if self.client.enable_crc:
                parts = sorted(self.finished_parts, key=lambda p: p.part_number)
                download_crc = cal_crc_from_upload_parts(parts)
                check_crc("upload_file", download_crc, result.hash_crc64_ecma, result.request_id)
            return result
        except Exception as e:
            raise TaskCompleteMultipartError(e)

    def _finish_part(self, part_info):
        self._callback_part_success(part_info)
        with self.lock:
            self.finished_parts.append(part_info)
            self.record["parts_info"].append({"part_number": part_info.part_number, "part_size": part_info.part_size,
                                              "offset": part_info.offset, "etag": part_info.etag,
                                              "hash_crc64ecma": part_info.hash_crc64_ecma, "is_completed": True})
            self.store.put(self.bucket, self.key, self.record)

    def _callback_success(self):
        self.event_listener(UploadEventType.Upload_Event_Complete_Multipart_Upload_Succeed)

    def _callback_fail(self, e):
        self.event_listener(UploadEventType.Upload_Event_Complete_Multipart_Upload_Failed, e)

    def _callback_part_success(self, part_info):
        self.event_listener(UploadEventType.Upload_Event_Upload_Part_Succeed, part_info=part_info)

    def _callback_part_fail(self, e, part):
        self.event_listener(UploadEventType.Upload_Event_Upload_Part_Failed, e)

    def _callback_abort(self, e):
        self.event_listener(UploadEventType.Upload_Event_UploadPart_Aborted, e)


class _BreakpointResumableCopyObject(BreakpointBase):
    def __init__(self, client, bucket, key, store: CheckPointStore, task_num,
                 parts_to_update, upload_id, record: Dict,
                 size, ssec_algorithm, ssec_key, ssec_key_md5, rate_limiter, cancel_hook,
                 datatransfer_listener, upload_event_listener,
                 src_bucket, src_object,
                 copy_source_if_match, copy_source_if_modified_since,
                 copy_source_if_none_match, copy_source_if_unmodified_since, src_version_id,
                 copy_source_ssec_algorithm, copy_source_ssec_key, copy_source_ssec_key_md5):
        super(_BreakpointResumableCopyObject, self).__init__(client=client, bucket=bucket, key=key, store=store,
                                                             task_num=task_num,
                                                             parts_to_do=parts_to_update, record=record, size=size,
                                                             rate_limiter=rate_limiter, cancel_hook=cancel_hook,
                                                             datatransfer_listener=datatransfer_listener,
                                                             event_listener=upload_event_listener)
        self.upload_id = upload_id
        self.src_version_id = src_version_id
        self.ssec_algorithm = ssec_algorithm
        self.ssec_key = ssec_key
        self.ssec_key_md5 = ssec_key_md5
        self.src_bucket = src_bucket
        self.src_object = src_object
        self.copy_source_if_match = copy_source_if_match
        self.copy_source_if_modified_since = copy_source_if_modified_since
        self.copy_source_if_none_match = copy_source_if_none_match
        self.copy_source_if_unmodified_since = copy_source_if_unmodified_since
        self.copy_source_ssec_algorithm = copy_source_ssec_algorithm
        self.copy_source_ssec_key = copy_source_ssec_key
        self.copy_source_ssec_key_md5 = copy_source_ssec_key_md5

    def _cover_to_finished_parts(self):
        for p in self.record["parts_info"]:
            self.finished_parts.append(
                PartInfo(part_number=p['part_number'], part_size=p['part_size'], offset=p['offset'], etag=p['etag'],
                         hash_crc64_ecma=p['hash_crc64ecma'], is_completed=True))

    def _do_task(self, part):
        try:
            result = self.client.upload_part_copy(self.bucket, self.key, self.upload_id, part.part_number,
                                                  self.src_bucket, self.src_object,
                                                  src_version_id=self.src_version_id,
                                                  copy_source_range_start=part.start,
                                                  # 由于拷贝为闭区间，因此end -1
                                                  copy_source_range_end=part.start + part.size - 1,
                                                  copy_source_if_match=self.copy_source_if_match,
                                                  copy_source_ssec_key=self.copy_source_ssec_key,
                                                  copy_source_ssec_algorithm=self.copy_source_ssec_algorithm,
                                                  copy_source_ssec_key_md5=self.copy_source_ssec_key_md5,
                                                  ssec_key=self.ssec_key,
                                                  ssec_algorithm=self.ssec_algorithm,
                                                  ssec_key_md5=self.ssec_key_md5)
        except Exception as e:
            self._callback_part_fail(e, part)
            raise e
        self._finish_part(
            PartInfo(part_number=part.part_number, part_size=part.size, offset=part.start, etag=result.etag,
                     hash_crc64_ecma=result.etag, is_completed=True))

    def _delete_checkpoint(self):
        with self.lock:
            self.store.delete(self.bucket, self.key, self.src_bucket, self.src_object, self.src_version_id)

    def _last_task(self, **kwargs):
        try:
            parts = _cover_to_uploaded_parts(self.finished_parts)
            result = self.client.complete_multipart_upload(self.bucket, self.key, self.upload_id, parts=parts)
            return result
        except Exception as e:
            raise TaskCompleteMultipartError(e)

    def _abort_task(self):
        self.client.abort_multipart_upload(self.bucket, self.key, self.upload_id)

    def _finish_part(self, part_info):
        self._callback_part_success(part_info)
        with self.lock:
            self.finished_parts.append(part_info)
            self.record["parts_info"].append({"part_number": part_info.part_number, "part_size": part_info.part_size,
                                              "offset": part_info.offset, "etag": part_info.etag,
                                              "hash_crc64ecma": part_info.hash_crc64_ecma, "is_completed": True})
            self.store.put(self.bucket, self.key, self.record, src_bucket=self.src_bucket, src_key=self.src_object,
                           version_id=self.src_version_id)

    def _callback_success(self):
        self.event_listener(CopyEventType.Copy_Event_Completed_Multipart_Upload_Succeed)

    def _callback_fail(self, e):
        self.event_listener(CopyEventType.Copy_Event_Completed_Multipart_Upload_Failed, e)

    def _callback_part_success(self, part_info):
        self.event_listener(CopyEventType.Copy_Event_Create_Part_Copy_Succeed,
                            part_info=CopyPartInfo(part_info.part_number,
                                                   part_info.offset,
                                                   part_info.offset + part_info.part_size - 1,
                                                   part_info.etag))

    def _callback_part_fail(self, e, part_info):
        self.event_listener(CopyEventType.Copy_Event_Create_Part_Copy_Failed, e,
                            CopyPartInfo(part_info.part_number,
                                         part_info.start,
                                         part_info.start + part_info.size - 1))

    def _callback_abort(self, e):
        self.event_listener(CopyEventType.Copy_Event_Create_Part_Copy_Aborted, e)


class _BreakpointDownloader(BreakpointBase):
    def __init__(self, client, bucket, key, file_path: str, store: CheckPointStore, task_num: int,
                 parts_to_download, record: Dict, etag,
                 datatransfer_listener, download_event_listener,
                 rate_limiter, cancel_hook, version_id, size,
                 ssec_algorithm, ssec_key, ssec_key_md5):
        super(_BreakpointDownloader, self).__init__(client=client, bucket=bucket, key=key, store=store,
                                                    task_num=task_num,
                                                    parts_to_do=parts_to_download, record=record, size=size,
                                                    rate_limiter=rate_limiter, cancel_hook=cancel_hook,
                                                    datatransfer_listener=datatransfer_listener,
                                                    event_listener=download_event_listener)
        self.etag = etag
        self.version_id = version_id
        self.file_path = file_path
        self.rate_limiter = rate_limiter
        self.ssec_algorithm = ssec_algorithm
        self.ssec_key = ssec_key
        self.ssec_key_md5 = ssec_key_md5
        self.temp = file_path + '.temp'

    def _finish_part(self, part):
        self._callback_part_success(part)
        with self.lock:
            self.finished_parts.append(part)
            self.record['parts_info'].append(
                {"part_number": part.part_number, "range_start": part.start, "range_end": part.end,
                 "hash_crc64ecma": part.part_crc, "is_completed": True})
            self.store.put(bucket=self.bucket, key=self.key, value=self.record, version_id=self.version_id)

    def _cover_to_finished_parts(self):
        self.finished_parts = []
        for p in self.record["parts_info"]:
            self.finished_parts.append(
                _PartToDo(p["part_number"], p["range_start"], p["range_end"], p["hash_crc64ecma"]))

    def _do_task(self, part):
        with open(self.temp, 'rb+') as f:
            try:
                f.seek(part.start, os.SEEK_SET)
                content = self.client.get_object(bucket=self.bucket, key=self.key, range_start=part.start,
                                                 range_end=part.end - 1, if_match=self.etag,
                                                 data_transfer_listener=self.datatransfer_listener,
                                                 rate_limiter=self.rate_limiter,
                                                 ssec_algorithm=self.ssec_algorithm,
                                                 ssec_key=self.ssec_key,
                                                 ssec_key_md5=self.ssec_key_md5,
                                                 version_id=self.version_id)
                copy_and_verify_length(content, f, part.end - part.start, request_id=content.request_id)
                if self.client.enable_crc:
                    part.part_crc = content.content.crc
            except Exception as e:
                self._callback_part_fail(e, part)
                raise e
            self._finish_part(part)

    def _pre_task(self):
        try:
            open(self.temp, 'a').close()
            self.event_listener(DownloadEventType.Download_Event_Create_TempFile_Succeed)
        except Exception as e:
            self.event_listener(DownloadEventType.Download_Event_Create_Temp_File_Failed, e)
            raise e

    def _delete_checkpoint(self):
        self.store.delete(bucket=self.bucket, key=self.key, version_id=self.version_id)

    def _last_task(self, **kwargs):
        try:
            tos_crc = kwargs.get('tos_crc')
            if self.client.enable_crc:
                parts = sorted(self.finished_parts, key=lambda p: p.part_number)
                download_crc = cal_crc_from_download_parts(parts)
                check_crc("download_file", download_crc, tos_crc, "")

            rename_file(self.temp, self.file_path)
        except Exception as e:
            raise RenameFileError(e)

    def _abort_task(self):
        # 删除临时文件
        os.remove(self.file_path)

    def _callback_success(self):
        self.event_listener(DownloadEventType.Download_Event_Rename_Temp_File_Succeed)

    def _callback_fail(self, e):
        self.event_listener(DownloadEventType.Download_Event_Rename_Temp_File_Failed, e)

    def _callback_abort(self, e):
        self.event_listener(DownloadEventType.Download_Event_Download_Part_Aborted, e)

    def _callback_rename_fail(self, e):
        self.event_listener(DownloadEventType.Download_Event_Rename_Temp_File_Failed, e)

    def _callback_part_success(self, part_info):
        self.event_listener(DownloadEventType.Download_Event_Download_Part_Succeed, part_info=part_info)

    def _callback_part_fail(self, e, part):
        self.event_listener(DownloadEventType.Download_Event_Download_Part_Failed, e, part)


class TaskExecutor(object):
    def __init__(self, num_task, task_fun, cancel_hook: CancelHook):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=num_task)
        self.task_fun = task_fun
        self.tasks = []
        self.cancel_hook = cancel_hook
        self._futures = []
        self._exc = None

    def submit(self, *args):
        future = self.executor.submit(self.task_fun, *args)
        self._futures.append(future)

    def run(self):
        try:
            # 运行前判定cancel
            if self.cancel_hook:
                self.cancel_hook.is_cancel()
            for future in as_completed(self._futures):
                e = future.exception()
                # 运行中判定cancel
                if self.cancel_hook:
                    self.cancel_hook.is_cancel()
                # 若为 TosServerError
                if isinstance(e, TosServerError):
                    self._exc = e
                    if _need_shutdown(e.status_code):
                        break
                # 其余异常直接抛出
                if e:
                    self._exc = e
                    break

            for future in self._futures:
                future.cancel()
            self.executor.shutdown(wait=True)
            if self._exc:
                raise self._exc
        finally:
            for future in self._futures:
                future.cancel()
            self.executor.shutdown(wait=True)


def _need_shutdown(code):
    return code in [403, 404, 405]



import json
from requests.structures import CaseInsensitiveDict

from . import exceptions
from .consts import CHUNK_SIZE
from .exceptions import TosClientError
from .utils import get_value, to_bytes


class Request(object):
    def __init__(self, method, url, path, host, data=None, params=None, headers=None):
        self.method = method
        self.url = url
        self.path = path
        self.data = to_bytes(data)
        self.params = params or {}
        self.headers = headers or {}

        self.headers['Host'] = host


class Response(object):
    def __init__(self, resp):
        self.resp = resp
        self.status = resp.status_code
        self.headers = CaseInsensitiveDict(resp.headers)
        self.content_length = get_value(self.headers, "content-length", lambda x: int(x))
        self.request_id = self.headers.get('x-tos-request-id', '')
        self._all_read = False
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(CHUNK_SIZE)
        if content:
            return content
        else:
            raise StopIteration

    def read(self, amt=None):
        if self._all_read:
            return b''

        if amt is None:
            content_list = []
            for chunk in self.resp.iter_content(CHUNK_SIZE):
                content_list.append(chunk)
            content = b''.join(content_list)

            self._all_read = True
            if self.content_length and len(content) != self.content_length:
                raise exceptions.TosClientError('IO Content not equal content-length')
            return content
        else:
            try:
                read = next(self.resp.iter_content(amt))
                self.offset += len(read)
                return read
            except StopIteration:
                if self.content_length and self.offset != self.content_length:
                    raise exceptions.TosClientError('IO Content not equal content-length')
                self._all_read = True
                return b''

    def json_read(self):
        try:
            body = self.read()
            return json.loads(body.decode('utf-8'))
        except Exception as e:
            raise TosClientError('unable to do serialization', e)



import logging
import threading
import time

logger = logging.getLogger(__name__)

DEFAULT_FETCH_TIME = 5 * 60


class Credentials():
    def __init__(self, access_key_id, access_key_secret, security_token=None):
        self.access_key_id = access_key_id.strip()
        self.access_key_secret = access_key_secret.strip()
        self.security_token = security_token

    def get_access_key_id(self):
        return self.access_key_id

    def get_access_key_secret(self):
        return self.access_key_secret

    def get_security_token(self):
        return self.security_token


class CredentialsProvider():
    def get_credentials(self):
        return

    def copy(self):
        return


class StaticCredentials(CredentialsProvider):
    def __init__(self, access_key_id, access_key_secret, security_token=None):
        self.credentials = Credentials(access_key_id, access_key_secret, security_token)

    def get_credentials(self):
        return self.credentials

    def copy(self):
        return Credentials(self.credentials.access_key_id, self.credentials.access_key_secret,
                           self.credentials.security_token)


class FederationToken():
    def __init__(self, access_key_id, access_key_secret, security_token, expiration, pre_fetch_sec=DEFAULT_FETCH_TIME):
        self.credential = Credentials(access_key_id, access_key_secret, security_token)
        self.expiration = expiration
        self.pre_fetch_sec = pre_fetch_sec

    def get_credentials(self):
        return self.credential

    def will_soon_expire(self):
        now = int(time.time())
        return now + self.pre_fetch_sec - self.expiration > 0

    def expire(self):
        return int(time.time()) > self.expiration


class FederationCredentials(CredentialsProvider):
    def __init__(self, get_credentials_func):
        self.get_credentials_func = get_credentials_func
        self.federationToken = None
        self.refreshing = 0
        self.__lock = threading.Lock()

    def get_credentials(self):
        # 不存在或者已经过期直接获取token
        if self.federationToken is None or self.federationToken.expire():
            return self._try_get_credential()
        # 快要过期且没有其他正在获取token的任务时，尝试去获取token
        if self.federationToken.will_soon_expire() and self.refreshing == 0:
            return self._try_get_credential()
        return self.federationToken.get_credentials()

    def _try_get_credential(self):
        with self.__lock:
            try:
                self.refreshing = 1
                # 再判断一次，因为可能已经被更新过了
                if self.federationToken is None or self.federationToken.will_soon_expire():
                    self.federationToken = self.get_credentials_func()
            except Exception as e:
                logger.error("get_credentials error: {0}".format(e))
                if self.federationToken is None:
                    raise
            finally:
                self.refreshing = 0
        return self.federationToken.get_credentials()



# coding: utf-8

import io
import os

from setuptools import find_packages, setup

# 包元信息
NAME = 'tos'  #
DESCRIPTION = 'Volc TOS (Tinder Object Storage) SDK'
URL = 'https://www.volcengine.com/'  #
EMAIL = 'sunyushan.jason@bytedance.com'
AUTHOR = 'sunyushan'

# 项目运行需要的依赖
REQUIRES = [
    'Deprecated>=1.2.13,<2.0.0',
    'pytz',
    "requests>=2.19.1,<2.26.0",
    'crcmod>=1.7',
    'six',
]

# 开发、测试过程中需要的依赖
DEV_REQUIRES = [
    'flake8>=3.5.0,<4.0.0',
    'mypy>=0.620; python_version>="3.4"',
    'tox>=3.0.0,<4.0.0',
    'isort>=4.0.0,<5.0.0',
    'pytest>=4.0.0,<5.0.0',
]

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except IOError:
    long_description = DESCRIPTION

about = {}
with io.open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)

setup(
    name=NAME,  # add the 'byted' prefix for package name
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='boilerplate',
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=REQUIRES,
    package_data={
        # for PEP484 & PEP561
        NAME: ['py.typed', '*.pyi'],
    },
)



# -*- coding: utf-8 -*-
import base64
import hashlib
import json
import logging
import warnings
from datetime import datetime
from typing import IO, Dict, Union

import requests
from deprecated import deprecated
from requests.adapters import HTTPAdapter
from requests.structures import CaseInsensitiveDict

from . import __version__
from . import exceptions
from .consts import (CONNECT_TIMEOUT, GMT_DATE_FORMAT)
from .convertor import (convert_complete_multipart_upload_result,
                        convert_copy_object_result,
                        convert_create_multipart_upload_result,
                        convert_delete_objects_result,
                        convert_get_object_acl_result,
                        convert_list_buckets_result,
                        convert_list_multipart_uploads_result,
                        convert_list_object_versions_result,
                        convert_list_objects_result, convert_list_parts_result,
                        convert_upload_part_copy_result)
from .http import Request, Response
from .models import (AppendObjectResult, CreateBucketResult, GetObjectResult,
                     HeadBucketResult, HeadObjectResult, PutObjectResult,
                     RequestResult)
from .utils import get_content_type, get_value, to_bytes, to_str, _format_endpoint, _get_host, _get_scheme, _if_map, \
    _make_virtual_host_uri, _get_virtual_host, _make_virtual_host_url, _cal_content_sha256

logger = logging.getLogger(__name__)

USER_AGENT = 'volc-tos-sdk-python/{0}'.format(__version__)



class TosClient():
    def __init__(self, auth, endpoint, connect_timeout=None, connection_pool_size=10, recognize_content_type=True):
        self.auth = auth
        self.endpoint = _format_endpoint(_if_map(auth.region, endpoint))
        self.host = _get_host(self.endpoint)
        self.scheme = _get_scheme(self.endpoint)
        self.timeout = connect_timeout or CONNECT_TIMEOUT
        self.recognize_content_type = recognize_content_type

        self.session = requests.Session()
        self.session.mount('http://', HTTPAdapter(pool_connections=connection_pool_size,
                                                  pool_maxsize=connection_pool_size))
        self.session.mount('https://', HTTPAdapter(pool_connections=connection_pool_size,
                                                   pool_maxsize=connection_pool_size))

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def generate_presigned_url(self, Method: str, Bucket: str = None, Key: str = None, Params: Dict = None,
                               ExpiresIn: int = None):
        """
        生成带签名的url
        :param: Bucket: 桶名
        :param: Key: 对象名
        :param: Params: 需要签名的HTTP查询参数
        :param: ExpiresIn: 过期时间，单位秒, 默认为1小时（3600秒）
        :param: HttpMethod: HTTP 方法， 如'GET', 'PUT' 等
        :return: 带签名的url
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        key = to_str(Key)
        params = Params or {}
        req = Request(
            Method,
            self._make_virtual_host_url(Bucket, key),
            _make_virtual_host_uri(key),
            _get_virtual_host(Bucket, self.endpoint),
            params=params,
        )
        return self.auth.sign_url(req, ExpiresIn)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def create_bucket(self, Bucket: str, ACL: str = None, GrantFullControl: str = None, GrantRead: str = None,
                      GrantReadACP: str = None, GrantWrite: str = None, GrantWriteACP: str = None):
        """
        创建bucket
        :param Bucket: 桶名
        :param ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param GrantFullControl: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWrite: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'

        :return: CreateBucketResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if ACL:
            headers['x-tos-acl'] = ACL

        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWrite:
            headers['x-tos-grant-write'] = GrantWrite
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        resp = self._req(bucket=Bucket, method='PUT', headers=headers)
        logger.info(
            'create_bucket, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id, resp.status))
        return CreateBucketResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def delete_bucket(self, Bucket: str):
        """
        删除指定bucket，bucket内不能有对象及分片数据
        :param Bucket: 桶名

        :return: RequestResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(bucket=Bucket, method='DELETE')
        logger.info(
            'delete_bucket, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id, resp.status))
        return RequestResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def head_bucket(self, Bucket: str):
        """
        获取指定bucket信息
        :param Bucket: 桶名
        :return: Response
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(bucket=Bucket, method='HEAD')
        logger.info(
            'head_bucket, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id, resp.status))
        return HeadBucketResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def list_buckets(self):
        """
        列举用户的bucket

        :return: ListBucketResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(method='GET')
        logger.info(
            'list_buckets, req id: {0}, status code: {1}'.format(resp.request_id, resp.status))
        return convert_list_buckets_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def list_objects(self, Bucket: str, Delimiter: str = None, EncodingType: str = None, Marker: str = None,
                     MaxKeys: int = None, Prefix: str = None):
        """
        列举对象
        :param Bucket: 桶名
        :param Delimiter: 目录分隔符
        :param EncodingType: 返回key编码类型
        :param Marker: 分页标志
        :param MaxKeys: 最大返回数
        :param Prefix: 前缀

        :return: ListObjectsResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {}
        if Delimiter:
            params['delimiter'] = Delimiter
        if EncodingType:
            params['encoding-type'] = EncodingType
        if MaxKeys:
            params['max-keys'] = MaxKeys
        if Prefix:
            params['prefix'] = Prefix
        if Marker:
            params['marker'] = Marker

        resp = self._req(bucket=Bucket, method='GET', params=params)
        logger.info(
            'list_objects, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id, resp.status))
        return convert_list_objects_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def list_object_versions(self, Bucket: str, Delimiter: str = None, EncodingType: str = None, KeyMarker: str = None,
                             MaxKeys: int = None, Prefix: str = None, VersionIdMarker: str = None):
        """
        列举多版本对象
        :param Bucket: 桶名
        :param Delimiter: 分隔符
        :param EncodingType: 返回key编码类型
        :param KeyMarker: 分页标志
        :param MaxKeys: 最大返回值
        :param Prefix: 前缀
        :param VersionIdMarker: 版本号分页标志
        :return: ListObjectVersionsResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'versions': ''}
        if Delimiter:
            params['delimiter'] = Delimiter
        if EncodingType:
            params['encoding-type'] = EncodingType
        if MaxKeys:
            params['max-keys'] = MaxKeys
        if Prefix:
            params['prefix'] = Prefix
        if KeyMarker:
            params['key-marker'] = KeyMarker
        if VersionIdMarker:
            params['version-id-marker'] = VersionIdMarker

        resp = self._req(bucket=Bucket, method='GET', params=params)
        logger.info(
            'list_object_versions, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id,
                                                                                      resp.status))
        return convert_list_object_versions_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def list_multipart_uploads(self, Bucket: str, Delimiter: str = None, EncodingType: str = None,
                               KeyMarker: str = None, MaxUploads: int = None, Prefix: str = None,
                               UploadIdMarker: str = None):
        """
        列举正在进行中的分片任务
        :param Bucket: 桶名
        :param Delimiter: 目录分隔符
        :param EncodingType: 'url'
        :param KeyMarker: 'string'
        :param MaxUploads: 一次列举的最大数量
        :param Prefix: 前缀
        :param UploadIdMarker: 分页标记

        :return: ListMultipartUploadsResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'uploads': ''}
        if Delimiter:
            params['delimiter'] = Delimiter
        if EncodingType:
            params['encoding-type'] = EncodingType
        if MaxUploads:
            params['max-uploads'] = MaxUploads
        if Prefix:
            params['prefix'] = Prefix
        if KeyMarker:
            params['key-marker'] = KeyMarker
        if UploadIdMarker:
            params['upload-id-marker'] = UploadIdMarker

        resp = self._req(bucket=Bucket, method='GET', params=params)
        logger.info(
            'list_multipart_uploads, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id,
                                                                                        resp.status))
        return convert_list_multipart_uploads_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def list_parts(self, Bucket: str, Key: str, UploadId: str, MaxParts: int = None, PartNumberMarker: int = None):
        """
        列举已经上传的分片
        :param Bucket: 桶名
        :param Key: 对象名
        :param UploadId: 分片任务id
        :param MaxParts: 一次列举最大分片数
        :param PartNumberMarker: 分页标记

        :return: ListPartsResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'uploadId': UploadId}
        if MaxParts:
            params['max-parts'] = MaxParts
        if PartNumberMarker:
            params['part-number-marker'] = PartNumberMarker

        resp = self._req(bucket=Bucket, key=Key, method='GET', params=params)
        logger.info('list_parts, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(
            Bucket, Key, resp.request_id, resp.status))
        return convert_list_parts_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def abort_multipart_upload(self, Bucket: str, Key: str, UploadId: str):
        """
        取消分片上传
        :param Bucket: 桶名
        :param Key: 对象名
        :param UploadId: 分片任务id

        :return: RequestResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(bucket=Bucket, key=Key, method='DELETE', params={'uploadId': UploadId})
        logger.info(
            'abort_multipart_upload, bucket: {0}, key: {1}, uploadId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, UploadId, resp.request_id, resp.status))
        return RequestResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def create_multipart_upload(self, Bucket: str, Key: str, ACL: str = None, CacheControl: str = None,
                                ContentDisposition: str = None, ContentEncoding: str = None,
                                ContentLanguage: str = None, ContentType: str = None, Expires: datetime = None,
                                GrantFullControl: str = None, GrantRead: str = None, GrantReadACP: str = None,
                                GrantWriteACP: str = None, Metadata: Dict = None, SSECustomerAlgorithm: str = None,
                                SSECustomerKey: str = None, SSECustomerKeyMD5: str = None):
        """
        初始化分片上传
        :param Bucket: 桶名
        :param Key: 对象名
        :param ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param CacheControl: 缓存控制
        :param ContentDisposition: 展示形式
        :param ContentEncoding: 报文编码
        :param ContentLanguage: 报文语言
        :param ContentType: 数据类型
        :param Expires: datetime(2021, 1, 1)
        :param GrantFullControl:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param Metadata: 自定义元数据
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: CreateMultipartUploadResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if Metadata:
            for k in Metadata:
                headers['x-tos-meta-' + k] = Metadata[k]

        if ACL:
            headers['x-tos-acl'] = ACL
        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        if CacheControl:
            headers['cache-control'] = CacheControl
        if ContentDisposition:
            headers['content-disposition'] = ContentDisposition
        if ContentEncoding:
            headers['content-encoding'] = ContentEncoding
        if ContentLanguage:
            headers['content-language'] = ContentLanguage

        if ContentType:
            headers['content-type'] = ContentType
        elif self.recognize_content_type:
            headers['content-type'] = get_content_type(Key)

        if Expires:
            headers['expires'] = Expires.strftime(GMT_DATE_FORMAT)

        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        resp = self._req(bucket=Bucket, key=Key, method='POST', params={'uploads': ''}, headers=headers)
        logger.info(
            'create_multipart_upload, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(
                Bucket, Key, resp.request_id, resp.status))
        return convert_create_multipart_upload_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def upload_part(self, Bucket: str, Key: str, PartNumber: int, UploadId: str, Body: Union[bytes, IO] = None):
        """
        上传分片
        :param Bucket: 桶名
        :param Key: 对象名
        :param PartNumber: 分片号，最小值为1
        :param UploadId: 分片任务id
        :param Body: 数据

        :return: PutObjectResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(bucket=Bucket, key=Key, method='PUT', params={'uploadId': UploadId, 'partNumber': PartNumber},
                         data=Body)
        logger.info(
            'upload_part, bucket: {0}, key: {1}, uploadId: {2}, partNumber: {3}, req id: {4}, status code: {5}'.format(
                Bucket, Key, UploadId, PartNumber, resp.request_id, resp.status))
        return PutObjectResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def upload_part_copy(self, Bucket: str, CopySource: Union[str, Dict], Key: str, PartNumber: int, UploadId: str,
                         CopySourceIfMatch: str = None, CopySourceIfModifiedSince: datetime = None,
                         CopySourceIfNoneMatch: str = None, CopySourceIfUnmodifiedSince: datetime = None,
                         CopySourceRange: str = None):
        """
        分片拷贝，把一个已有文件拷贝成目标文件的一个分片
        :param Bucket: 目标桶名
        :param CopySource: 源桶名 'string' or {'Bucket': 'string', 'Key': 'string', 'VersionId': 'string'}
        :param Key: 目标对象名
        :param PartNumber: 分片号， 最小值为1
        :param UploadId: 分片任务id
        :param CopySourceIfMatch:
        :param CopySourceIfModifiedSince:
        :param CopySourceIfNoneMatch:
        :param CopySourceIfUnmodifiedSince:
        :param CopySourceRange:

        :return: UploadPartCopyResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if isinstance(CopySource, str):
            headers['x-tos-copy-source'] = CopySource
        elif isinstance(CopySource, dict):
            copy_source = CopySource['Bucket'] + '/' + CopySource['Key']
            if 'VersionId' in CopySource:
                copy_source = copy_source + '?versionId=' + CopySource['VersionId']
            headers['x-tos-copy-source'] = copy_source

        if CopySourceIfMatch:
            headers['x-tos-copy-source-if-match'] = CopySourceIfMatch
        if CopySourceIfModifiedSince:
            headers['x-tos-copy-source-if-modified-since'] = CopySourceIfModifiedSince.strftime(GMT_DATE_FORMAT)
        if CopySourceIfNoneMatch:
            headers['x-tos-copy-source-if-none-match'] = CopySourceIfNoneMatch
        if CopySourceIfUnmodifiedSince:
            headers['x-tos-copy-source-if-unmodified-since'] = CopySourceIfUnmodifiedSince.strftime(GMT_DATE_FORMAT)

        if CopySourceRange:
            headers['x-tos-copy-source-range'] = CopySourceRange

        resp = self._req(bucket=Bucket, key=Key, method='PUT', params={'uploadId': UploadId, 'partNumber': PartNumber},
                         headers=headers)
        logger.info(
            'upload_part_copy, bucket: {0}, key: {1}, uploadId: {2}, partNumber: {3}, req id: {4}, status code: {5}'.format(
                Bucket, Key, UploadId, PartNumber, resp.request_id, resp.status))
        return convert_upload_part_copy_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def complete_multipart_upload(self, Bucket: str, Key: str, UploadId: str, MultipartUpload: Dict = None):
        """
        合并分片
        :param Bucket: 桶名
        :param Key: 对象名
        :param UploadId: 分片任务id
        :param MultipartUpload: {
        'Parts': [
            {
                'ETag': 'string',
                'PartNumber': 123
            },
        ]
        }

        :return: CompleteMultipartUploadResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        data = json.dumps(MultipartUpload)
        resp = self._req(bucket=Bucket, key=Key, method='POST', params={'uploadId': UploadId}, data=data)
        logger.info(
            'complete_multipart_upload, bucket: {0}, key: {1}, uploadId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, UploadId, resp.request_id, resp.status))
        return convert_complete_multipart_upload_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def put_object(self, Bucket: str, Key: str, ACL: str = None, Body: Union[bytes, IO] = None,
                   CacheControl: str = None, ContentDisposition: str = None, ContentEncoding: str = None,
                   ContentMD5: str = None, ContentLanguage: str = None, ContentType: str = None,
                   Expires: datetime = None,
                   GrantFullControl: str = None, GrantRead: str = None, GrantReadACP: str = None,
                   GrantWriteACP: str = None, Metadata: Dict = None, SSECustomerAlgorithm: str = None,
                   SSECustomerKey: str = None, SSECustomerKeyMD5: str = None):
        """
        上传对象
        :param Bucket: 桶名
        :param Key: 对象名
        :param ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param Body: 数据
        :param CacheControl: 缓存控制
        :param ContentDisposition: 展示形式
        :param ContentEncoding: 报文编码
        :param ContentLanguage: 报文语言
        :param ContentType: 数据类型
        :param Expires: datetime(2021, 1, 1)
        :param GrantFullControl:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param Metadata: 自定义元数据
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: PutObjectResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if Metadata:
            for k in Metadata:
                headers['x-tos-meta-' + k] = Metadata[k]
        if ACL:
            headers['x-tos-acl'] = ACL

        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        if ContentMD5:
            headers['Content-MD5'] = ContentMD5

        if CacheControl:
            headers['cache-control'] = CacheControl
        if ContentDisposition:
            headers['content-disposition'] = ContentDisposition
        if ContentEncoding:
            headers['content-encoding'] = ContentEncoding
        if ContentLanguage:
            headers['content-language'] = ContentLanguage

        if ContentType:
            headers['content-type'] = ContentType
        elif self.recognize_content_type:
            headers['content-type'] = get_content_type(Key)

        if Expires:
            headers['expires'] = Expires.strftime(GMT_DATE_FORMAT)

        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        resp = self._req(bucket=Bucket, key=Key, method='PUT', data=Body, headers=headers)
        logger.info(
            'put_object, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(
                Bucket, Key, resp.request_id, resp.status))
        return PutObjectResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def get_object(self, Bucket: str, Key: str, IfMatch: str = None, IfModifiedSince: datetime = None,
                   IfNoneMatch: str = None, IfUnmodifiedSince: datetime = None, Range: str = None,
                   ResponseCacheControl: str = None, ResponseContentDisposition: str = None,
                   ResponseContentEncoding: str = None, ResponseContentLanguage: str = None,
                   ResponseContentType: str = None, ResponseExpires: datetime = None, VersionId: str = None,
                   SSECustomerAlgorithm: str = None, SSECustomerKey: str = None, SSECustomerKeyMD5: str = None):
        """
        下载对象
        :param Bucket: 桶名
        :param Key: 对象名
        :param IfMatch: 只有在匹配时，才返回对象
        :param IfModifiedSince: datetime(2021, 1, 1)
        :param IfNoneMatch: 只有在不匹配时，才返回对象
        :param IfUnmodifiedSince: datetime(2021, 1, 1)
        :param Range: 下载范围
        :param ResponseCacheControl: 指定回包的Cache-Control
        :param ResponseContentDisposition: 指定回包的Content-Disposition
        :param ResponseContentEncoding: 指定回包的Content-Encoding
        :param ResponseContentLanguage: 指定回包的Content-Language
        :param ResponseContentType: 指定回包的Content-Type
        :param ResponseExpires: 指定回包的Expires
        :param VersionId: 版本号
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: RequestResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if IfMatch:
            headers['If-Match'] = IfMatch
        if IfModifiedSince:
            headers['If-Modified-Since'] = IfModifiedSince.strftime(GMT_DATE_FORMAT)
        if IfNoneMatch:
            headers['If-None-Match'] = IfNoneMatch
        if IfUnmodifiedSince:
            headers['If-Unmodified-Since'] = IfUnmodifiedSince.strftime(GMT_DATE_FORMAT)

        if Range:
            headers['Range'] = Range

        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        params = {}
        if VersionId:
            params['versionId'] = VersionId
        if ResponseCacheControl:
            params['response-cache-control'] = ResponseCacheControl
        if ResponseContentDisposition:
            params['response-content-disposition'] = ResponseContentDisposition
        if ResponseContentEncoding:
            params['response-content-encoding'] = ResponseContentEncoding
        if ResponseContentLanguage:
            params['response-content-language'] = ResponseContentLanguage
        if ResponseContentType:
            params['response-content-type'] = ResponseContentType
        if ResponseExpires:
            params['response-expires'] = ResponseExpires.strftime(GMT_DATE_FORMAT)

        resp = self._req(bucket=Bucket, key=Key, method='GET', headers=headers, params=params)
        logger.info(
            'get_object, bucket: {0}, key: {1}, VersionId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, VersionId, resp.request_id, resp.status))
        return GetObjectResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def head_object(self, Bucket: str, Key: str, IfMatch: str = None, IfModifiedSince: datetime = None,
                    IfNoneMatch: str = None, IfUnmodifiedSince: datetime = None, Range: str = None,
                    VersionId: str = None, SSECustomerAlgorithm: str = None, SSECustomerKey: str = None,
                    SSECustomerKeyMD5: str = None):
        """
        查看对象信息
        :param Bucket: 桶名
        :param Key: 对象名
        :param IfMatch: 只有在匹配时，才返回对象
        :param IfModifiedSince: datetime(2021, 1, 1)
        :param IfNoneMatch: 只有在不匹配时，才返回对象
        :param IfUnmodifiedSince: datetime(2021, 1, 1)
        :param Range: 下载范围
        :param VersionId: 版本号
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: HeadObjectResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}
        if IfMatch:
            headers['If-Match'] = IfMatch
        if IfModifiedSince:
            headers['If-Modified-Since'] = IfModifiedSince.strftime(GMT_DATE_FORMAT)
        if IfNoneMatch:
            headers['If-None-Match'] = IfNoneMatch
        if IfUnmodifiedSince:
            headers['If-Unmodified-Since'] = IfUnmodifiedSince.strftime(GMT_DATE_FORMAT)
        if Range:
            headers['Range'] = Range
        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        params = {}
        if VersionId:
            params['versionId'] = VersionId
        resp = self._req(bucket=Bucket, key=Key, method='HEAD', params=params, headers=headers)
        logger.info(
            'head_object, bucket: {0}, key: {1}, versionId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, VersionId, resp.request_id, resp.status))
        return HeadObjectResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def delete_object(self, Bucket: str, Key: str, VersionId: str = None):
        """
        删除对象
        :param Bucket: 桶名
        :param Key: 对象名
        :param VersionId: 版本号

        :return: RequestResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {}
        if VersionId:
            params = {'versionId': VersionId}
        resp = self._req(bucket=Bucket, key=Key, params=params, method='DELETE')
        logger.info(
            'delete_object, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(
                Bucket, Key, resp.request_id, resp.status))
        return RequestResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def copy_object(self, Bucket: str, CopySource: Union[str, Dict], Key: str, ACL: str = None,
                    CacheControl: str = None,
                    ContentDisposition: str = None, ContentEncoding: str = None, ContentLanguage: str = None,
                    ContentType: str = None, CopySourceIfMatch: str = None, CopySourceIfModifiedSince: datetime = None,
                    CopySourceIfNoneMatch: str = None, CopySourceIfUnmodifiedSince: datetime = None,
                    Expires: datetime = None, GrantFullControl: str = None, GrantRead: str = None,
                    GrantReadACP: str = None, GrantWriteACP: str = None, Metadata: Dict = None,
                    MetadataDirective: str = None, SSECustomerAlgorithm: str = None, SSECustomerKey: str = None,
                    SSECustomerKeyMD5: str = None):
        """
        拷贝对象
        :param Bucket: 目标桶名
        :param CopySource:
        :param Key: 目标对象名
        :param ACL: private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param CacheControl: 缓存控制
        :param ContentDisposition: 展示形式
        :param ContentEncoding: 报文编码
        :param ContentLanguage: 报文语言
        :param ContentType: 数据类型
        :param CopySourceIfMatch: 源对象匹配时，才返回对象
        :param CopySourceIfModifiedSince: datetime(2021, 1, 1)
        :param CopySourceIfNoneMatch: 源对象不匹配时，才返回对象
        :param CopySourceIfUnmodifiedSince: datetime(2021, 1, 1)
        :param Expires: datetime(2021, 1, 1)
        :param GrantFullControl:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param Metadata: 自定义元数据
        :param MetadataDirective: 'COPY'|'REPLACE'
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: CopyObjectResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        headers = {}

        if isinstance(CopySource, str):
            headers['x-tos-copy-source'] = CopySource
        elif isinstance(CopySource, dict):
            copy_source = CopySource['Bucket'] + '/' + CopySource['Key']
            if 'VersionId' in CopySource:
                copy_source = copy_source + '?versionId=' + CopySource['VersionId']
            headers['x-tos-copy-source'] = copy_source

        if MetadataDirective:
            headers['x-tos-metadata-directive'] = MetadataDirective

        if Metadata:
            for k in Metadata:
                headers['x-tos-meta-' + k] = Metadata[k]

        if CacheControl:
            headers['cache-control'] = CacheControl
        if ContentDisposition:
            headers['content-disposition'] = ContentDisposition
        if ContentEncoding:
            headers['content-encoding'] = ContentEncoding
        if ContentLanguage:
            headers['content-language'] = ContentLanguage
        if ContentType:
            headers['content-type'] = ContentType
        if Expires:
            headers['expires'] = Expires.strftime(GMT_DATE_FORMAT)

        if ACL:
            headers['x-tos-acl'] = ACL
        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        if CopySourceIfMatch:
            headers['x-tos-copy-source-if-match'] = CopySourceIfMatch
        if CopySourceIfModifiedSince:
            headers['x-tos-copy-source-if-modified-since'] = CopySourceIfModifiedSince.strftime(GMT_DATE_FORMAT)
        if CopySourceIfNoneMatch:
            headers['x-tos-copy-source-if-none-match'] = CopySourceIfNoneMatch
        if CopySourceIfUnmodifiedSince:
            headers['x-tos-copy-source-if-unmodified-since'] = CopySourceIfUnmodifiedSince.strftime(GMT_DATE_FORMAT)

        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        resp = self._req(bucket=Bucket, key=Key, method='PUT', headers=headers)
        logger.info(
            'copy_object, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(Bucket, Key, resp.request_id,
                                                                                       resp.status))
        return convert_copy_object_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def append_object(self, Bucket: str, Key: str, Offset: str, ACL: str = None, Body: Union[bytes, IO] = None,
                      CacheControl: str = None, ContentDisposition: str = None, ContentEncoding: str = None,
                      ContentMD5: str = None, ContentLanguage: str = None, ContentType: str = None,
                      Expires: datetime = None,
                      GrantFullControl: str = None, GrantRead: str = None, GrantReadACP: str = None,
                      GrantWriteACP: str = None, Metadata: Dict = None, SSECustomerAlgorithm: str = None,
                      SSECustomerKey: str = None, SSECustomerKeyMD5: str = None):
        """
        追加写对象
        :param Bucket: 桶名
        :param Key: 对象名
        :param ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param Body: 数据
        :param CacheControl: 缓存控制
        :param ContentDisposition: 展示形式
        :param ContentEncoding: 报文编码
        :param ContentLanguage: 报文语言
        :param ContentType: 数据类型
        :param Expires: datetime(2021, 1, 1)
        :param GrantFullControl:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param Metadata: 自定义元数据
        :param SSECustomerAlgorithm: 'AES256'
        :param SSECustomerKey: 加密密钥
        :param SSECustomerKeyMD5: 密钥md5值

        :return: AppanedObjectResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'append': ''}
        params['offset'] = Offset
        headers = {}
        if Metadata:
            for k in Metadata:
                headers['x-tos-meta-' + k] = Metadata[k]
        if ACL:
            headers['x-tos-acl'] = ACL

        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        if ContentMD5:
            headers['Content-MD5'] = ContentMD5

        if CacheControl:
            headers['cache-control'] = CacheControl
        if ContentDisposition:
            headers['content-disposition'] = ContentDisposition
        if ContentEncoding:
            headers['content-encoding'] = ContentEncoding
        if ContentLanguage:
            headers['content-language'] = ContentLanguage

        if ContentType:
            headers['content-type'] = ContentType
        elif self.recognize_content_type:
            headers['content-type'] = get_content_type(Key)

        if Expires:
            headers['expires'] = Expires.strftime(GMT_DATE_FORMAT)

        if SSECustomerAlgorithm:
            headers['x-tos-server-side-encryption-customer-algorithm'] = SSECustomerAlgorithm
        if SSECustomerKey:
            headers['x-tos-server-side-encryption-customer-key'] = SSECustomerKey
        if SSECustomerKeyMD5:
            headers['x-tos-server-side-encryption-customer-key-md5'] = SSECustomerKeyMD5

        resp = self._req(bucket=Bucket, key=Key, method='POST', data=Body, headers=headers, params=params)
        logger.info(
            'append_object, bucket: {0}, key: {1}, req id: {2}, status code: {3}'.format(
                Bucket, Key, resp.request_id, resp.status))
        return AppendObjectResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def put_object_acl(self, Bucket: str, Key: str, ACL: str = None, AccessControlPolicy: Dict = None,
                       GrantFullControl: str = None, GrantRead: str = None, GrantReadACP: str = None,
                       GrantWrite: str = None, GrantWriteACP: str = None, VersionId: str = None):
        """
        设置对象acl
        :param Bucket: 桶名
        :param Key: 对象名
        :param ACL: 'private'|'public-read'|'public-read-write'|'authenticated-read'|'bucket-owner-read'|
        'bucket-owner-full-control'
        :param AccessControlPolicy:
        {
            "Grants": [
                {
                    "Grantee": {
                        "ID": "string",
                        "Type": "CanonicalUser"|"Group",
                        "Canned": "AuthenticatedUsers"|"AllUsers"
                    },
                    "Permission": "FULL_CONTROL"|"WRITE"|"WRITE_ACP"|"READ"|"READ_ACP"
                },
            ],
            "Owner":{
                "ID":"xxx",
                "DisplayName":"xxx"
            }
        }
        :param GrantFullControl:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantRead:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantReadACP:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWrite:  'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param GrantWriteACP: 'id="xxx",canned="AllUsers"|"AuthenticatedUsers"'
        :param VersionId: 版本号

        :return: RequestResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'acl': ''}
        if VersionId:
            params['versionId'] = VersionId

        headers = {}
        if ACL:
            headers['x-tos-acl'] = ACL

        if GrantFullControl:
            headers['x-tos-grant-full-control'] = GrantFullControl
        if GrantRead:
            headers['x-tos-grant-read'] = GrantRead
        if GrantReadACP:
            headers['x-tos-grant-read-acp'] = GrantReadACP
        if GrantWrite:
            headers['x-tos-grant-write'] = GrantWrite
        if GrantWriteACP:
            headers['x-tos-grant-write-acp'] = GrantWriteACP

        data = None
        if AccessControlPolicy:
            data = json.dumps(AccessControlPolicy)

        resp = self._req(bucket=Bucket, key=Key, method='PUT', params=params, headers=headers, data=data)
        logger.info(
            'put_object_acl, bucket: {0}, key: {1}, versionId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, VersionId, resp.request_id, resp.status))
        return RequestResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def get_object_acl(self, Bucket: str, Key: str, VersionId: str = None):
        """
        获取对象的acl
        :param Bucket: 桶名
        :param Key: 对象名
        :param VersionId: 版本号
        :return: GetObjectAclResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        params = {'acl': ''}
        if VersionId:
            params['versionId'] = VersionId
        resp = self._req(bucket=Bucket, key=Key, method='GET', params=params)
        logger.info(
            'get_object_acl, bucket: {0}, key: {1}, versionId: {2}, req id: {3}, status code: {4}'.format(
                Bucket, Key, VersionId, resp.request_id, resp.status))
        return convert_get_object_acl_result(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def set_object_metadata(self, Bucket: str, Key: str, ObjectMetadata: Dict = {}):
        """
        设置对象元数据
        :param Bucket: 桶名
        :param Key: 对象名
        :param ObjectMetadata: 要修改的元数据
        :return:
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        resp = self._req(bucket=Bucket, key=Key, method='Post', headers=ObjectMetadata, params={'metadata': ''})
        logger.info(
            'set_object_metadata, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id,
                                                                                     resp.status))
        return RequestResult(resp)

    @deprecated(version='2.1.0', reason="please use TosClientV2")
    def delete_objects(self, Bucket: str, Delete: Dict):
        """
        批量删除对象
        :param Bucket: 桶名
        :param Delete: {
        'Objects': [
            {
                'Key': 'string',
                'VersionId': 'string'
            },
        ],
        'Quiet': True|False
        }

        :return: DeleteObjectsResult
        """
        warnings.warn("please use TosClientV2", DeprecationWarning)
        data = json.dumps(Delete)

        headers = {}
        headers['Content-MD5'] = to_str(base64.b64encode(hashlib.md5(to_bytes(data)).digest()))
        resp = self._req(bucket=Bucket, method='POST', data=data, headers=headers, params={'delete': ''})
        logger.info(
            'delete_objects, bucket: {0}, req id: {1}, status code: {2}'.format(Bucket, resp.request_id, resp.status))
        return convert_delete_objects_result(resp)

    def _make_virtual_host_url(self, bucket=None, key=None):
        return _make_virtual_host_url(self.host, self.scheme, bucket, key)

    def _req(self, bucket=None, key=None, method=None, data=None, headers=None, params=None):
        key = to_str(key)
        data = to_bytes(data)

        headers = CaseInsensitiveDict(headers)

        if headers.get('x-tos-content-sha256') is None:
            headers['x-tos-content-sha256'] = _cal_content_sha256(data)

        req = Request(method, self._make_virtual_host_url(bucket, key),
                      _make_virtual_host_uri(key),
                      _get_virtual_host(bucket, self.endpoint),
                      data=data,
                      params=params,
                      headers=headers)
        self.auth.sign_request(req)

        if 'User-Agent' not in req.headers:
            req.headers['User-Agent'] = USER_AGENT
        res = self.session.request(method,
                                   req.url,
                                   data=req.data,
                                   headers=req.headers,
                                   params=req.params,
                                   stream=True,
                                   timeout=self.timeout)

        rsp = Response(res)
        if rsp.status >= 300:
            e = exceptions.make_exception(rsp)
            logger.info('Exception: %s' % e)
            raise e

        content_length = get_value(rsp.headers, 'content-length', int)
        if content_length is not None and content_length == 0:
            rsp.read()

        return rsp



import json

_TOS_ERROR_TO_EXCEPTION = {}


def get_value(kv, key, handler=lambda x: x):
    if key in kv:
        return handler(kv[key])
    else:
        return None


class TosError(Exception):
    def __init__(self, msg: str):
        self.message = msg

    def __str__(self):
        error = {"message": self.message}
        return str(error)


def make_exception(resp):
    details = _parse_error_body(resp)
    return TosError(details.get('Message', ''))


def _parse_error_body(resp):
    body = resp.read()
    try:
        return json.loads(body.decode('utf-8'))
    except Exception:
        return {'message': body.decode('utf-8')}


class TosClientError(TosError):
    def __init__(self, msg: str, cause: Exception = None):
        super(TosClientError, self).__init__(msg)
        self.cause = cause

    def __str__(self):
        error = {'message': self.message,
                 'case': str(self.cause)}
        return str(error)


class TosServerError(TosError):
    def __init__(self, resp, msg: str, code: str, host_id: str, resource: str):
        self.message = msg
        self.request_id = resp.request_id
        self.id2 = get_value(resp.headers, "x-tos-id-2")
        self.status_code = resp.status
        self.header = resp.headers
        self.code = code
        self.host_id = host_id
        self.resource = resource

    def __str__(self):
        error = {'message': self.message,
                 'request_id': self.request_id,
                 'id2': self.id2,
                 'status_code': self.status_code,
                 'header': self.header,
                 'code': self.code,
                 'host_id': self.host_id,
                 'resource': self.resource}
        return str(error)


def make_server_error(resp):
    details = _parse_error_body(resp)
    code = details.get('Code', '')
    host_id = details.get('HostId', '')
    resource = details.get('Resource', '')
    message = details.get('Message', '')
    return TosServerError(resp, message, code, host_id, resource)


def make_server_error_with_exception(resp, body):
    code = body.get('Code', '')
    host_id = body.get('HostId', '')
    resource = body.get('Resource', '')
    message = body.get('Message', '')
    return TosServerError(resp, message, code, host_id, resource)


class CancelError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        info = {'message': self.message}
        return str(info)


class CancelWithAbortError(CancelError):
    def __init__(self, message):
        super(CancelWithAbortError, self).__init__(message)


class CancelNotWithAbortError(CancelError):
    def __init__(self, message):
        super(CancelNotWithAbortError, self).__init__(message)


class TaskCompleteMultipartError(Exception):
    def __init__(self, cause):
        self.cause = cause
        self.message = 'failed to do complete multipart task '

    def __str__(self):
        error = {'message': self.message,
                 'case': str(self.cause)}
        return str(error)


class RenameFileError(Exception):
    def __init__(self, cause):
        self.cause = cause
        self.message = 'failed to do rename file task'

    def __str__(self):
        error = {'message': self.message,
                 'case': str(self.cause)}
        return str(error)
